#include  "axis.h"
#include  <cstdlib>
#include  <memory> 
#include  "lexer.h" 
#include  "parsetree.h" 
#include  <algorithm>
#include  "atlas-types.h"
#include  "bitmap.h"

namespace atlas { namespace interpreter {

volatile std::sig_atomic_t interrupt_flag=0;

layer::list layer::lexical_context;

type_expr last_type;
shared_value last_value;

namespace {

class frame
{
  const id_pat& pattern;
public:
  static shared_context current;

  frame (const id_pat& pattern)
  : pattern(pattern)
  { assert(count_identifiers(pattern)>0); // avoid frames without identifiers
    current = std::make_shared<evaluation_context>(std::move(current));
  }
  ~frame() { current = current->tail(); } // don't use |std::move| here!

  void bind (const shared_value& val)
    { current->reserve(count_identifiers(pattern));
      thread_components(pattern,val,current->back_inserter());
    }
  std::vector<id_type> id_list() const; // list identifiers, for back-tracing
};

class lambda_frame
{
  const id_pat& pattern;
  const shared_context saved;
  const bool empty;
public:
  lambda_frame (const id_pat& pattern, const shared_context& outer)
  : pattern(pattern)
  , saved(std::move(frame::current))
  , empty(count_identifiers(pattern)==0)
  { assert(&outer!=&frame::current); // for excluded case use |frame| instead
    try {
      frame::current =
        empty ? outer : std::make_shared<evaluation_context>(outer);
    }
    catch(...)
    { frame::current = std::move(saved); throw; }
    // restore as destructor would do
  }
  ~lambda_frame() { frame::current = std::move(saved); }

  bool is_empty() const { return empty; }
  void bind (const shared_value& val);
  std::vector<id_type> id_list() const; // list identifiers, for back-tracing
};

class dest_iterator
{
  multiple_assignment::loc_list::weak_const_iterator l_it;
  multiple_assignment::glob_list::weak_const_iterator g_it;
  const BitMap& is_global; 
  unsigned long n; // index into |is_global|
 public:
  dest_iterator 
   (const multiple_assignment::loc_list& locs
   ,const multiple_assignment::glob_list& globs
   ,const BitMap& is_global)
  : l_it(locs.wcbegin())
  , g_it(globs.wcbegin())
  , is_global(is_global)
  , n(0)
  {}
  void receive (const shared_value& val);
};

struct threader
{ typedef containers::sl_list<multiple_assignment::local_dest> loc_list;
  typedef containers::sl_list<shared_share> glob_list;

  const expr& e;
  loc_list locs;
  glob_list globs;
  BitMap is_global;
  containers::sl_list<std::pair<id_type,const_type_p> > assoc;

  threader (const expr& e) : e(e), locs(), globs(), is_global(), assoc() {}
  void thread (const id_pat& pat,type_expr& type); // recursively analyse |pat|
  void refine () const; // maybe specialise stored identifiers
};


shared_context frame::current; 

bool while_condition_result;


void check_interrupt()
{
  if (interrupt_flag!=0)
  { interrupt_flag=0;
    throw user_interrupt();
  }
}

void balance
   ( type_expr& target // component type required from context
   , raw_expr_list elist // list of expression to be balanced
   , const expr& e // containing expression, for error reporting
   , const char* description // what kind of components, for error message
   , std::vector<expression_ptr>& components // output, converted expressions
   )
{
  unsigned n = length(elist); components.reserve(n);
  std::vector<type_expr> comp_type; comp_type.reserve(n);
  type_expr common;
    // greatest common denominator that branch types convert to\dots
  containers::sl_list<type_expr> conflicts;
    // except those branch types that are put aside here
  
  for (wel_const_iterator it(elist); not it.at_end(); ++it)
  { try
    { comp_type.push_back(target.copy());
          // start each with a copy of original |target| type
      type_expr& ctype =comp_type.back(); // call that copy |ctype|
      components.push_back(expression_ptr());
         // push, whether or not |convert_expr| succeeds
      components.back()=convert_expr(*it,ctype);
      if (not broader_eq(common,ctype))
        { if (broader_eq(ctype,common))
            common = ctype.copy();
          else
            conflicts.push_back(ctype.copy());
            // record type not convertible to |common|
        }
    }
    catch (balance_error& err)
    { if (&err.offender!=&*it) // only incorporate top-level balancing errors
        throw; // any deeper error is propagated to be reported
      else if (err.offender.kind==list_display)
        // then wrap variants in row-of
        for (auto it=err.variants.wbegin(); not err.variants.at_end(it); ++it)
          it->set_from(type_expr(type_ptr(new type_expr(std::move(*it)))));
          // row-of |*it|
      conflicts.append(std::move(err.variants)); // then join to our |conflicts|
    }
  }
  
  { for (auto it=conflicts.cbegin(); not conflicts.at_end(it); )
      if (broader_eq(common,*it))
        conflicts.erase(it);
      else
        ++it;
    if (conflicts.empty())
       // then balancing succeeded, so set |target| to |common|
    { bool success = target.specialise(common); ndebug_use(success);
      assert(success);
      // since |common| was obtained by |convert_expr| from some copy of |target|
    }
    else
    { balance_error err(e);
      (err.message += " between ") += description ;
      err.variants.push_back(std::move(common));
      err.variants.append(std::move(conflicts));
      throw std::move(err);
    }
  }
  wel_const_iterator it(elist);
  for (unsigned i=0; i<n; ++i,++it)
    if (comp_type[i]!=target)
      components[i] = convert_expr(*it,target);
      // redo conversion with broader |common| type
}

id_type size_of_name()
{ static id_type name=main_hash_table->match_literal("#");
  return name;
}
id_type concatenate_name()
{ static id_type name=main_hash_table->match_literal("##");
  return name;
}
id_type print_name()
{ static id_type name=main_hash_table->match_literal("print");
  return name;
}
id_type to_string_name()
{ static id_type name=main_hash_table->match_literal("to_string");
  return name;
}
id_type prints_name()
{ static id_type name=main_hash_table->match_literal("prints");
  return name;
}
id_type error_name()
{ static id_type name=main_hash_table->match_literal("error");
  return name;
}

inline bool is_special_operator(id_type id)
{ return id==size_of_name()
      or id==concatenate_name()
      or id==print_name()
      or id==to_string_name()
      or id==prints_name()
      or id==error_name(); }

id_type equals_name()
{ static id_type name=main_hash_table->match_literal("=");
  return name;
}

expression_ptr make_variadic_call
  (const shared_variadic_builtin& f
  ,const std::string& name
  ,expression_ptr&& a, bool needs_voiding
  ,const source_location& loc)
{ if (needs_voiding)
    a.reset(new voiding(std::move(a))); // wrap argument in voiding
  return expression_ptr(new variadic_builtin_call(f,name,std::move(a),loc));
}

void extend_message
  (error_base& e,const call_base* call, const shared_function& f,
   const std::string& arg)
{ std::ostringstream o;
  o << "\n(in call " << call->loc << " of " << call->function_name() << ", ";
  f->report_origin(o);
  o << ')';
  if (verbosity>0)
    o << "\n  argument" << (arg[0]=='(' ? "s: " : ": ") << arg;
  e.message.append(o.str());
}

bool is_pair_type(const type_expr& t)
{ return t.kind==tuple_type and length(t.tupple)==2; }

id_pat copy_id_pat(const id_pat& p)
{
  containers::sl_list<id_pat> rl;
  std::transform(p.sublist.begin(),end(p.sublist)
                , std::back_inserter(rl), copy_id_pat);
  return id_pat (p.name, p.kind, rl.undress());
}

std::vector<id_type> frame::id_list() const
{ std::vector<id_type> names; names.reserve(count_identifiers(pattern));
  list_identifiers(pattern,names);
  return names;
}

void lambda_frame::bind (const shared_value& val)
{ assert(not empty); // one should not call |bind| for an |empty| lambda frame
  check_interrupt();
  frame::current->reserve(count_identifiers(pattern));
  thread_components(pattern,val,frame::current->back_inserter());
}

std::vector<id_type> lambda_frame::id_list() const
{ std::vector<id_type> names; names.reserve(count_identifiers(pattern));
  list_identifiers(pattern,names);
  return names;
}

template <  template < unsigned > class  slice  >
expression make_slice(unsigned flags
  ,expression_ptr&& a, expression_ptr&& b, expression_ptr&& c)
{ switch (flags)
  {
  case 0: return new slice<0>(std::move(a),std::move(b),std::move(c));
  case 1: return new slice<1>(std::move(a),std::move(b),std::move(c));
  case 2: return new slice<2>(std::move(a),std::move(b),std::move(c));
  case 3: return new slice<3>(std::move(a),std::move(b),std::move(c));
  case 4: return new slice<4>(std::move(a),std::move(b),std::move(c));
  case 5: return new slice<5>(std::move(a),std::move(b),std::move(c));
  case 6: return new slice<6>(std::move(a),std::move(b),std::move(c));
  case 7: return new slice<7>(std::move(a),std::move(b),std::move(c));
  default: assert(false); return(nullptr);
  }
}

bool was_negated (expr& e, bool b)
{ switch (e.kind)
  { default: return b;
  case negation_expr:
    e=std::move(*e.negation_variant); // remove top level |not|
    return was_negated(e,not b); // recursively continue
  case let_expr: return was_negated(e.let_variant->body,b);
  case seq_expr: return was_negated(e.sequence_variant->last,b);
  case next_expr: return was_negated(e.sequence_variant->first,b);
  }
}

bool was_negated(expr& e) { return was_negated(e,false); }

type_ptr unknown_union_type(size_t length)
{ type_list l; auto n = std::max<size_t>(2,length);
  while (n-->0)
    l.push_front(unknown_type.copy());
  return mk_union_type(std::move(l));
}

constexpr bool in_reversed(unsigned flags) { return (flags&0x1)!=0; }
constexpr bool in_forward(unsigned flags) { return (flags&0x1)==0; }
constexpr bool out_reversed(unsigned flags) { return (flags&0x2)!=0; }
constexpr bool out_forward(unsigned flags) { return (flags&0x2)==0; }
constexpr bool no_frame(unsigned flags) { return (flags&0x4)!=0; }
constexpr bool has_frame(unsigned flags) { return (flags&0x4)==0; }
constexpr bool yields_count(unsigned flags) { return (flags&0x8)!=0; }

expression make_while_loop (unsigned flags,expression_ptr&& body)
{ switch(flags)
  {
  case 0x0: return new while_expression<0>(std::move(body));
  case 0x2: return new while_expression<2>(std::move(body));
  case 0x8: return new while_expression<8>(std::move(body));
  default: assert(false); return nullptr;
  }
}

expression make_for_loop
  (unsigned flags, const id_pat& id, expression_ptr&& i, expression_ptr&& b
  ,subscr_base::sub_type t)
{ switch(t)
  {
  case subscr_base::row_entry:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::row_entry>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::row_entry>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::row_entry>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::row_entry>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::vector_entry:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::vector_entry>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::vector_entry>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::vector_entry>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::vector_entry>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::ratvec_entry:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::ratvec_entry>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::ratvec_entry>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::ratvec_entry>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::ratvec_entry>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::string_char:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::string_char>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::string_char>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::string_char>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::string_char>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::matrix_column:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::matrix_column>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::matrix_column>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::matrix_column>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::matrix_column>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  case subscr_base::mod_poly_term:
    switch (flags)
    {
    case 0: return new for_expression<0,subscr_base::mod_poly_term>
      (id,std::move(i),std::move(b));
    case 1: return new for_expression<1,subscr_base::mod_poly_term>
      (id,std::move(i),std::move(b));
    case 2: return new for_expression<2,subscr_base::mod_poly_term>
      (id,std::move(i),std::move(b));
    case 3: return new for_expression<3,subscr_base::mod_poly_term>
      (id,std::move(i),std::move(b));
    default: assert(false); return(nullptr);
    }
  default: assert(false); return(nullptr);
  }
}

expression make_counted_loop (unsigned flags, id_type id, 
   expression_ptr&& count, expression_ptr&& bound, expression_ptr&& body)
{ switch(flags)
  {
  case 0: return new  counted_for_expression<0>
    (id,std::move(count),std::move(bound),std::move(body));
  case 1: return new  counted_for_expression<1>
    (id,std::move(count),std::move(bound),std::move(body));
  case 2: return new  counted_for_expression<2>
    (id,std::move(count),std::move(bound),std::move(body));
  case 3: return new  counted_for_expression<3>
    (id,std::move(count),std::move(bound),std::move(body));
   case 4: return new  counted_for_expression<4>
    (id,std::move(count),std::move(bound),std::move(body));
  case 6: return new  counted_for_expression<6>
    (id,std::move(count),std::move(bound),std::move(body));
  default: assert(false); return(nullptr);
  }
}

inline bool functype_specialise
  (type_expr& t, const type_expr& from, const type_expr& to)
{ return t.specialise(gen_func_type) 
  and t.func->arg_type.specialise(from) 
  and t.func->result_type.specialise(to);
}

void thread_assign
  (const id_pat& pat, const shared_value& val, dest_iterator& dst)
{ if ((pat.kind&0x2)!=0)
  { const tuple_value* t=force<tuple_value>(val.get());
    assert(t->val.size()==length(pat.sublist));
    auto src = t->val.begin();
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it,++src)
      thread_assign(*it,*src,dst);
  }
  if ((pat.kind&0x1)!=0)
    dst.receive(val);
}

void dest_iterator::receive (const shared_value& val)
{ if (is_global.isMember(n++))
  { assert (not g_it.at_end());
    *(*g_it++) = val;
  } // send to |shared_share| stored in |globs|
  else
  {
    assert (not l_it.at_end());
    frame::current->elem(l_it->depth,l_it->offset)=val;
    ++l_it;
  }
}

void report_undefined (id_type id,const expr& e,const char* where)
{ std::ostringstream o;
  o << "Undefined identifier '" << main_hash_table->name_of(id)
  << "' in " << where << ' ' << e;
  if (e.loc.file!=Hash_table::empty)
    o << ' ' << e.loc;
  throw program_error (o.str());
}

void report_constant_modified (id_type id,const expr& e,const char* where)
{ std::ostringstream o;
  o << "Name '" << main_hash_table->name_of(id)
    << "' is constant in " << where << ' ' << e;
  if (e.loc.file!=Hash_table::empty)
    o << ' ' << e.loc;
  throw program_error (o.str());
}

void print_wrapper(expression_base::level l)
{
  *output_stream << *execution_stack.back() << std::endl;
  if (l!=expression_base::single_value) // in |single_value| case we are done
    push_expanded(l,pop_value()); // otherwise remove and possibly expand value
}

std::ostream& to_string_aux(std::ostream& o, expression_base::level l)
{ shared_value v=pop_value();

  const string_value* s=dynamic_cast<const string_value*>(v.get());
  if (s!=nullptr)
    o << s->val; // single string without quotes
  else
  { const tuple_value* t=dynamic_cast<const tuple_value*>(v.get());
    if (t!=nullptr)
    { for (auto it=t->val.begin(); it!=t->val.end(); ++it)
      { s=dynamic_cast<const string_value*>(it->get());
        if (s!=nullptr)
	  o << s->val; // string components without quotes
        else
           o << *it->get(); // treat non-string tuple components as |print|
      }
    }
    else
      o << *v; // output like |print| unless string or tuple
  }
  return o;
}

void to_string_wrapper(expression_base::level l)
{ std::ostringstream o;
  to_string_aux(o,l);
  if (l!=expression_base::no_value)
    push_value(std::make_shared<string_value>(o.str()));
}

void prints_wrapper(expression_base::level l)
{ to_string_aux(*output_stream,l) << std::endl;
  if (l==expression_base::single_value)
    wrap_tuple<0>(); // don't forget to return a value if asked for
}

void error_wrapper(expression_base::level l)
{ std::ostringstream o;
  to_string_aux(o,l);
  throw runtime_error(o.str());
}

void sizeof_wrapper(expression_base::level l)
{ size_t s=get<row_value>()->val.size();
  if (l!=expression_base::no_value)
    push_value(std::make_shared<int_value>(s));
}

void suffix_element_wrapper(expression_base::level l)
{ shared_value e=pop_value();
  own_row r=get_own<row_value>();
  if (l!=expression_base::no_value)
  { r->val.push_back(std::move(e));
    push_value(std::move(r));
  }
}

void prefix_element_wrapper(expression_base::level l)
{ own_row r=get_own<row_value>();
  shared_value e=pop_value();
  if (l!=expression_base::no_value)
  { r->val.insert(r->val.begin(),e);
    push_value(std::move(r));
  }
}

void join_rows_wrapper(expression_base::level l)
{ shared_row second=get<row_value>();
  shared_row first=get<row_value>();
  if (l==expression_base::no_value)
    return;
  const auto& x=first->val;
  const auto& y=second->val;
  own_row result = std::make_shared<row_value>(x.size()+y.size());
std::copy(y.begin(),y.end(),
       std::copy(x.begin(),x.end(),result->val.begin())  );
push_value(std::move(result));
}


void join_rows_row_wrapper(expression_base::level l)
{ shared_row arg=get<row_value>();
  if (l==expression_base::no_value)
    return;
  const std::vector<shared_value>& x=arg->val;
  std::vector< const std::vector<shared_value>*> p; p.reserve(x.size());
  for (auto it=x.cbegin(); it!=x.cend(); ++it)
    p.push_back(&force<row_value>(it->get())->val);
  size_t s=0;
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    s+=(*it)->size();
  own_row result = std::make_shared<row_value>(s);
  auto dst=result->val.begin();
  for (auto it=p.cbegin(); it!=p.cend(); ++it)
    dst=std::copy((*it)->cbegin(),(*it)->cend(),dst);
  assert(dst==result->val.end());
push_value(std::move(result));
}

void bool_not_wrapper(expression_base::level l)
{ bool b=get<bool_value>()->val;
  if (l!=expression_base::no_value)
    push_value(whether(not b));
}


static shared_builtin sizeof_row_builtin =
    std::make_shared<const builtin_value<false> >(sizeof_wrapper,"#@[T]");
static shared_builtin sizeof_vector_builtin =
    std::make_shared<const builtin_value<false> >
      (sizeof_vector_wrapper,"#@vec");
static shared_builtin sizeof_ratvec_builtin =
    std::make_shared<const builtin_value<false> >
      (sizeof_ratvec_wrapper,"#@ratvec");
static shared_builtin sizeof_string_builtin =
    std::make_shared<const builtin_value<false> >
       (sizeof_string_wrapper,"#@string");
static shared_builtin matrix_columns_builtin =
    std::make_shared<const builtin_value<false> >
     (matrix_ncols_wrapper,"ncols@mat");
static shared_builtin sizeof_parampol_builtin =
    std::make_shared<const builtin_value<false> >
      (virtual_module_size_wrapper, "#@ParamPol");
static shared_variadic_builtin print_builtin =
  std::make_shared<const builtin_value<true> >(print_wrapper,"print@T");
static shared_variadic_builtin to_string_builtin =
  std::make_shared<const builtin_value<true> >(to_string_wrapper,"to_string@T");
static shared_variadic_builtin prints_builtin =
  std::make_shared<const builtin_value<true> >(prints_wrapper,"prints@T");
static shared_variadic_builtin error_builtin =
  std::make_shared<const builtin_value<true> >(error_wrapper,"error@T");
static shared_builtin prefix_elt_builtin =
  std::make_shared<const builtin_value<false> >
    (prefix_element_wrapper,"#@(T,[T])");
static shared_builtin suffix_elt_builtin =
  std::make_shared<const builtin_value<false> >
    (suffix_element_wrapper,"#@([T],T)");
static shared_builtin join_rows_builtin =
  std::make_shared<const builtin_value<false> >
    (join_rows_wrapper,"##@([T],[T])");
static shared_builtin join_rows_row_builtin =
  std::make_shared<const builtin_value<false> >
    (join_rows_row_wrapper,"##@([[T]])");
static shared_builtin boolean_negate_builtin =
  std::make_shared<const builtin_value<false> >(bool_not_wrapper,"not@bool");

}

void reset_evaluator()
{ if (!execution_stack.empty())
  { if (verbosity==0)
      execution_stack.clear();
    else
    {
      std::cerr << "Discarding from execution stack:" << std::endl;
      do
        std::cerr << *pop_value() << std::endl;
      while (!execution_stack.empty());
    }
  }
}

layer::layer(size_t n) // non-function non-loop layer
: variable(), constness(n)
, loop_depth(lexical_context.empty() ? 0 : lexical_context.front()->loop_depth)
, return_type(lexical_context.empty() ? nullptr
             : lexical_context.front()->return_type)
{ variable.reserve(n); lexical_context.push_front(this); }
layer::layer(size_t n,type_p return_type) // function or loop layer
: variable(), constness(n)
, loop_depth(return_type!=nullptr ? 0
            : lexical_context.empty() ? 1
            : lexical_context.front()->loop_depth+1)
, return_type(return_type!=nullptr ? return_type 
             : lexical_context.empty() ? nullptr
             : lexical_context.front()->return_type)
{ variable.reserve(n); lexical_context.push_front(this); }

void layer::add(id_type id,type_expr&& t,bool is_const)
{ for (auto it=variable.begin(); it!=variable.end(); ++it)
  // traverse |variable| vector
    if (it->first==id)
      throw program_error 
       (std::string("Multiple binding of '")
                    +main_hash_table->name_of(id)
                    +"' in same scope");
  constness.set_to(variable.size(),is_const);
  variable.emplace_back( id, std::move(t) );
}

const_type_p layer::lookup
  (id_type id, size_t& depth, size_t& offset, bool& is_const)
{ size_t i=0;
  for (auto range=lexical_context.cbegin(); not lexical_context.at_end(range);
       ++range)
    if (not (*range)->variable.empty())
    { for (auto it=(*range)->cbegin(); it!=(*range)->cend(); ++it)
        if (it->first==id) // then found; now set output values
          { depth=i;
            offset=it-(*range)->begin();
            is_const = (*range)->is_const(it);
            return &it->second;
          }
      ++i; // increment depth for non-empty layers only
    }
  return nullptr;
}

void layer::specialise (size_t depth, size_t offset,const type_expr& t)
{ auto range=lexical_context.cbegin();
  while ((*range)->variable.empty() or depth-->0)
    ++range;
  (**range)[offset].second.specialise(t);
}

expression_ptr convert_expr(const expr& e, type_expr& type)
{ switch(e.kind)
  {
   
   case integer_denotation:
     { expression_ptr d(new denotation
         (std::make_shared<int_value>(
            big_int(e.str_denotation_variant.c_str(),10))));
       return conform_types(int_type,type,std::move(d),e);
     }
   case string_denotation:
     { expression_ptr d(new denotation
         (std::make_shared<string_value>(e.str_denotation_variant)));
       return conform_types(str_type,type,std::move(d),e);
     }
   case boolean_denotation:
     { expression_ptr d(new denotation
           (whether(e.bool_denotation_variant)));
       return conform_types(bool_type,type,std::move(d),e);
     }
   
   case last_value_computed:
   { std::ostringstream o;
     o << '(' << last_type << ":$)"; 
   return conform_types
       (last_type
       ,type
       ,expression_ptr(new capture_expression(last_value,o.str()))
       ,e);
   }
   
   case die_expr:
   { return expression_ptr(new shell); }
   
   case break_expr:
   { if (layer::may_break(e.break_variant))
       return expression_ptr(new breaker(e.break_variant));
     throw expr_error(e,"not in the reach of a loop");
   }
   
   case return_expr:
   { if (layer::may_return())
       return expression_ptr(new 
         returner(convert_expr(*e.return_variant,layer::current_return_type())));
     throw expr_error(e,"not inside a function body");
   }
   
   case tuple_display:
   { type_expr tup=unknown_tuple(length(e.sublist));
     std::unique_ptr<tuple_expression> tup_exp(new tuple_expression(0));
     std::vector<expression_ptr>& comp = tup_exp->component;
     comp.reserve(length(e.sublist));
     bool tuple_expected = type.specialise(tup);
     // whether |type| is a tuple of correct size
     wtl_iterator tl_it (tuple_expected ? type.tupple : tup.tupple);
     for (wel_const_iterator it(e.sublist); not it.at_end(); ++it,++tl_it)
     { comp.push_back(convert_expr(*it,*tl_it));
       if (*tl_it==void_type and not is_empty(*it))
         comp.back().reset(new voiding(std::move(comp.back())));
     }
     expression_ptr result(std::move(tup_exp));
     if (tuple_expected or coerce(tup,type,result))
       return result;
     throw type_error(e,std::move(tup),std::move(type));
   }
   
   case list_display:
   { std::unique_ptr<list_expression> result (new list_expression(0));
     auto& comps = result->component;
     result->component.reserve(length(e.sublist));
   static const char* const str = "components of list expression";
     if (type.specialise(row_of_type))
     {
       balance(*type.component_type,e.sublist,e,str,comps);
       if (*type.component_type==void_type)
         
         { wel_const_iterator it(e.sublist);
           for (auto cit=comps.begin(); cit!=comps.end(); ++it,++cit)
             if (not is_empty(*it))
               cit->reset(new voiding(std::move(*cit)));
         }
       return std::move(result);
     }
   
     type_expr comp_type;
     if (type==void_type) // in void context leave undetermined target type
     { balance(comp_type,e.sublist,e,str,result->component);
       return std::move(result);
       // and forget |comp_type|
     }
   
     const conversion_record* conv = row_coercion(type,comp_type);
     if (conv!=nullptr)
     { balance(comp_type,e.sublist,e,str,result->component);
       return expression_ptr(new
         conversion(*conv,expression_ptr(std::move(result))));
     }
   
     throw type_error(e,row_of_type.copy(),std::move(type));
     // |type| incompatible with any list
   }
   
   case applied_identifier:
   { const id_type id=e.identifier_variant;
     const_type_p id_t; size_t i,j; bool is_const;
     const bool is_local=(id_t=layer::lookup(id,i,j,is_const))!=nullptr;
     if (not is_local and (id_t=global_id_table->type_of(id,is_const))==nullptr)
     {
       std::ostringstream o;
       o << "Undefined identifier '" << main_hash_table->name_of(id) << '\'';
       if (e.loc.file!=Hash_table::empty)
         o << ' ' << e.loc;
       throw program_error (o.str());
     }
   
     expression_ptr id_expr =  is_local
     ? expression_ptr(new local_identifier(id,i,j))
     : expression_ptr(new global_identifier(id));
     if (type.specialise(*id_t)) // then required type admits known identifier type
       { if (type!=*id_t)
         // usage has made type of identifier more specialised
         { if (is_local)
             layer::specialise(i,j,type);
             // then refine its type in the local context
           else
             global_id_table->specialise(id,type); // or in |global_id_table|
         }
         return id_expr;
       }
     else if (coerce(*id_t,type,id_expr))
       return id_expr;
     throw type_error(e,id_t->copy(),std::move(type));
   }
   
   case function_call:
   { const application_node& call = *e.call_variant;
     if (call.fun.kind==applied_identifier)
       
       { const id_type id =call.fun.identifier_variant;
         size_t i,j; bool b; // dummies; local binding not used here
         auto local_type_p=layer::lookup(id,i,j,b);
         if (local_type_p==nullptr or local_type_p->kind!=function_type)
        // not calling by local identifier
         { const overload_table::variant_list& variants
             = global_overload_table->variants(id);
           if (variants.size()>0 or is_special_operator(id))
           return resolve_overload(e,type,variants);
         }
       }
     type_expr f_type=gen_func_type.copy(); // start with generic function type
     expression_ptr fun = convert_expr(call.fun,f_type);
     expression_ptr arg = convert_expr(call.arg,f_type.func->arg_type);
     if (f_type.func->arg_type==void_type and not is_empty(call.arg))
       arg.reset(new voiding(std::move(arg)));
     expression_ptr result(new 
        call_expression(std::move(fun),std::move(arg),e.loc));
     return conform_types(f_type.func->result_type,type,std::move(result),e);
   }
   
   case let_expr:
   { const auto& lexp=*e.let_variant;
     const id_pat& pat=lexp.pattern;
     type_expr decl_type=pattern_type(pat);
     expression_ptr arg = convert_expr(lexp.val,decl_type);
   auto n=count_identifiers(pat);
     if (n==0)
     // then avoid frame without identifiers, so compile as sequence expression
       return expression_ptr(new 
         seq_expression(std::move(arg),convert_expr(lexp.body,type)));
     if (decl_type==void_type and not is_empty(lexp.val))
       // rare case, introducing void identifier
       arg.reset(new voiding(std::move(arg)));
     layer new_layer(n);
     thread_bindings(pat,decl_type,new_layer,false);
     return expression_ptr(new 
       let_expression(pat,std::move(arg),convert_expr(lexp.body,type)));
   }
   
   case lambda_expr:
   { const lambda_node& fun=*e.lambda_variant;
     const id_pat& pat=fun.pattern;
     const type_expr& arg_type=fun.parameter_type;
       // argument type specified in |fun|
     if (not pattern_type(pat).specialise(arg_type))
       // do |pat| structure and |arg_type| conflict?
       throw expr_error(e,"Function argument pattern does not match its type");
     type_expr* rt; type_expr dummy;
     if (type.specialise(gen_func_type)
                and type.func->arg_type.specialise(arg_type))
       rt = &type.func->result_type; // we can now safely access this
     else if (type==void_type)
       rt=&dummy; // in void context there is no return type to set
     else
       throw type_error(e,
                          type_expr(arg_type.copy(),unknown_type.copy()),
                          std::move(type));
   layer new_layer(count_identifiers(pat),rt);
     thread_bindings(pat,arg_type,new_layer,false);
   return expression_ptr(new 
         lambda_expression(pat, convert_expr(fun.body,*rt), std::move(e.loc)));
   }
   
   case seq_expr:
   { const sequence_node& seq=*e.sequence_variant;
     expression_ptr first = convert_expr(seq.first,as_lvalue(void_type.copy()));
     expression_ptr last  = convert_expr(seq.last,type);
     return expression_ptr(new seq_expression(std::move(first),std::move(last)));
   }
   break;
   case next_expr:
   { const sequence_node& seq=*e.sequence_variant;
     expression_ptr first = convert_expr(seq.first,type);
     expression_ptr last  = convert_expr(seq.last,as_lvalue(void_type.copy()));
     return expression_ptr(new next_expression(std::move(first),std::move(last)));
   }
   break;
   
   case subscription:
   { type_expr array_type, index_type, subscr_type;
       // all initialised to |undetermined_type|
     auto& subsn = *e.subscription_variant;
       // alias |struct| pointed to by the |unique_ptr|
     expression_ptr array = convert_expr(subsn.array,array_type);
     expression_ptr index = convert_expr(subsn.index,index_type);
     expression_ptr subscr;
     
     switch (subscr_base::index_kind(array_type,index_type,subscr_type))
     { case subscr_base::row_entry:
       if (subsn.reversed)
         subscr.reset(new
           row_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           row_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::vector_entry:
       if (subsn.reversed)
         subscr.reset(new
           vector_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           vector_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::ratvec_entry:
       if (subsn.reversed)
         subscr.reset(new
           ratvec_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           ratvec_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::string_char:
       if (subsn.reversed)
         subscr.reset(new
           string_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           string_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::matrix_entry:
       if (subsn.reversed)
         subscr.reset(new
           matrix_subscription<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           matrix_subscription<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::matrix_column:
       if (subsn.reversed)
         subscr.reset(new
           matrix_get_column<true>(std::move(array),std::move(index)));
       else
         subscr.reset(new
           matrix_get_column<false>(std::move(array),std::move(index)));
     break;
     case subscr_base::mod_poly_term:
       if (subsn.reversed)
         throw expr_error(e,"Cannot do reversed subscription of a ParamPol");
       subscr.reset(new module_coefficient(std::move(array),std::move(index)));
     break;
     case subscr_base::not_so:
       std::ostringstream o;
       o << "Cannot subscript value of type " << array_type 
         << " with index of type " << index_type;
       throw expr_error(e,o.str());
     }
     return conform_types(subscr_type,type,std::move(subscr),e);
   }
   
   case slice:
   { type_expr array_type; // initialised to |undetermined_type|
     expression_ptr array = convert_expr(e.slice_variant->array,array_type);
     expression_ptr lower =
       convert_expr(e.slice_variant->lower,as_lvalue(int_type.copy()));
     expression_ptr upper =
       convert_expr(e.slice_variant->upper,as_lvalue(int_type.copy()));
     expression_ptr subscr; const unsigned fl = e.slice_variant->flags.to_ulong();
     switch (subscr_base::slice_kind(array_type))
       { case subscr_base::row_entry: subscr.reset(
        make_slice<row_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       case subscr_base::vector_entry: subscr.reset(
        make_slice<vector_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       case subscr_base::ratvec_entry: subscr.reset(
        make_slice<ratvec_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       case subscr_base::string_char: subscr.reset(
        make_slice<string_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       case subscr_base::matrix_column: subscr.reset(
        make_slice<matrix_slice>
           (fl,std::move(array),std::move(lower),std::move(upper)));
       break;
       default: std::ostringstream o;
         o << "Cannot slice value of type " << array_type;
         throw expr_error(e,o.str());
       }
   
     return conform_types(array_type,type,std::move(subscr),e);
   }
   
   case negation_expr:
   { type_expr b=bool_type.copy();
      expression_ptr arg = convert_expr(*e.negation_variant,b);
     if (not type.specialise(b)) // |not| preserves the |bool| type
       throw type_error(e,std::move(b),std::move(type));
     return expression_ptr(new  builtin_call
        (boolean_negate_builtin,std::move(arg),e.loc));
   }
   
   case conditional_expr:
   { auto& exp = *e.if_variant;
     if (was_negated(exp.condition)) // eliminate negated conditions
       exp.branches.contents.swap(exp.branches.next->contents);
     expression_ptr c  =  convert_expr(exp.condition,as_lvalue(bool_type.copy()));
     std::vector<expression_ptr> conv;
     balance(type,&exp.branches,e,"branches of conditional",conv);
   return expression_ptr(new 
       conditional_expression(std::move(c),std::move(conv[0]),std::move(conv[1])));
   }
   
   case int_case_expr:
   { auto& exp = *e.if_variant;
     expression_ptr c  =  convert_expr(exp.condition,as_lvalue(int_type.copy()));
     std::vector<expression_ptr> conv;
     balance(type,&exp.branches,e,"branches of case",conv);
   return expression_ptr(new 
       int_case_expression(std::move(c),std::move(conv)));
   }
   
   case union_case_expr:
   { auto& exp = *e.if_variant;
     size_t n_branches=length(&exp.branches);
     type_expr switch_type;
     std::ostringstream o; // for use in various error messages
   
     expression_ptr c  =  convert_expr(exp.condition,switch_type);
     if (switch_type.kind!=union_type)
       throw type_error(e,std::move(switch_type),
                          std::move(*unknown_union_type(n_branches)));
   
     size_t n_variants=length(switch_type.tupple);
     if (n_variants!=n_branches)
       
       { o << "Union case expression has " << n_branches 
           << "branches,\nwhile the union type " << switch_type 
           << " has " << n_variants << " variants";
         throw expr_error(e,o.str());
       }
   
     std::vector<expression_ptr> branches;
   
     auto variant_type_p = switch_type.tupple;
     for (auto branch_p=&exp.branches; branch_p!=nullptr;
          branch_p=branch_p->next.get(), variant_type_p=variant_type_p->next.get())
     {
       auto f_type_p =
         mk_function_type(std::move(variant_type_p->contents),type.copy());
       auto branch_f = convert_expr(branch_p->contents,*f_type_p);
       branches.push_back(std::move(branch_f));
       type.specialise(f_type_p->func->result_type);
     }
   return expression_ptr(new 
       union_case_expression(std::move(c),std::move(branches)));
   }
   
   case discrimination_expr:
   { auto& exp = *e.disc_variant;
     size_t n_branches=length(&exp.branches);
     type_expr subject_type;
     std::ostringstream o; // for use in various error messages
   
     expression_ptr c  =  convert_expr(exp.subject,subject_type);
     if (subject_type.kind!=union_type)
       throw type_error(e,std::move(subject_type),
                          std::move(*unknown_union_type(n_branches)));
     const wtl_const_iterator types_start(subject_type.tupple);
   
     const auto type_name = typedef_table.find(subject_type);
     if (type_name==type_table::no_id)
       
       {
         o << "Discrimination on expression of type " << subject_type 
           << " requires naming its variants";
         throw expr_error(e,o.str());
       }
     const auto& field_names = typedef_table.fields(type_name);
   
     size_t n_variants=length(subject_type.tupple);
   
     std::vector<choice_part> choices(n_variants);
   
     
     { shared_expression default_choice;
       for (auto branch_p=&exp.branches; branch_p!=nullptr;
            branch_p=branch_p->next.get())
       { const auto& branch = branch_p->contents;
         if (branch.label==type_table::no_id)
           
           { if (default_choice.get()!=nullptr)
               throw expr_error(e,"Multiple default branches present");
             default_choice = convert_expr(branch.branch,type);
           }
         else
         { size_t k = std::find(field_names.begin(), field_names.end(),branch.label)
                      -field_names.begin();
           
           { if (k==field_names.size())
             { o << "Branch has label " << main_hash_table->name_of(branch.label) 
                 << " not associated to any variant of the union type "
                 << subject_type;
               throw expr_error(e,o.str());
             }
             if (choices[k].second.get()!=nullptr)
             { o << "Multiple branches with label "
                 << main_hash_table->name_of(branch.label);
               throw expr_error(e,o.str());
             }
           }
           const auto& variant_type = *std::next(types_start,k);
           // type of variant for this branch
         
         { if (not pattern_type(branch.pattern).specialise(variant_type))
           { o << "Pattern " << branch.pattern 
               << " does not match type " << variant_type 
               << " for variant " <<  main_hash_table->name_of(branch.label);
             throw expr_error(e,o.str());
           }
         
           auto n=count_identifiers(branch.pattern);
           expression_ptr result;
           layer branch_layer(n);
           thread_bindings(branch.pattern,variant_type,branch_layer,false);
           result=convert_expr(branch.branch,type);
         choices[k] = choice_part 
              (n==0 ? id_pat() : copy_id_pat(branch.pattern)
              ,std::move(result));
         }
         }
       }
       
       {
         if (n_branches>n_variants)
           throw expr_error(e,"Spurious default branch present");
         else if (n_branches<n_variants and default_choice.get()==nullptr)
         {
           auto it=choices.begin();
           while(it->second.get()!=nullptr) ++it;
           auto variant=field_names[it-choices.begin()];
           if (variant==type_table::no_id)
             o << "Missing branch for anonymous variant " << it-choices.begin();
           else
             o << "Missing branch for variant " << main_hash_table->name_of(variant);
           o << " of type " << subject_type << " in discrimination expression";
           throw expr_error(e,o.str());
         }
       
         if (default_choice.get()!=nullptr)
         // if a default was given, insert for omitted branches
           for (auto it=choices.begin(); it!=choices.end(); ++it)
             if (it->second.get()==nullptr)
               *it = choice_part(id_pat(),default_choice); // share |default_choice| here
       }
     }
   return expression_ptr(new 
       discrimination_expression(std::move(c),std::move(choices)));
   }
   
   case while_expr:
   { while_node& w=*e.while_variant;
     layer bind(0,nullptr); // no local variables for loop, but allow |break|
   
     if (type==void_type)
       return expression_ptr(make_while_loop (w.flags.to_ulong(),
          convert_expr(w.body, as_lvalue(void_type.copy()))));
     else if (type==int_type)
       return expression_ptr(make_while_loop  (0x8,
          convert_expr(w.body, as_lvalue(void_type.copy()))));
     else if (type.specialise(row_of_type))
     { auto& comp_type = *type.component_type;
       expression_ptr b = convert_expr(w.body,comp_type);
       if (comp_type==void_type and not is_empty(w.body))
         b.reset(new voiding(std::move(b)));
       return expression_ptr (make_while_loop 
          (w.flags.to_ulong(),std::move(b)));
     }
     else
     
     { type_expr comp_type;
       const conversion_record* conv = row_coercion(type,comp_type);
       if (conv==nullptr)
         throw type_error(e,row_of_type.copy(),std::move(type));
     
       return expression_ptr(new conversion(*conv, expression_ptr(make_while_loop 
            (w.flags.to_ulong(),convert_expr(w.body,comp_type)))));
     }
   }
   
   case do_expr:
   { sequence_node& seq=*e.sequence_variant;
     bool neg = was_negated(seq.first);
     expression_ptr body = convert_expr(seq.last,type);
       // body needs type-checking in all cases
     if (seq.first.kind==boolean_denotation)
     { if (seq.first.bool_denotation_variant==neg)
         return expression_ptr(new dont_expression()); // and drop |body|
       else
         return expression_ptr(new forever_expression(std::move(body)));
     }
     expression_ptr condition = convert_expr(seq.first,as_lvalue(bool_type.copy()));
     if (neg)
       return expression_ptr(new 
         do_expression<true>(std::move(condition),std::move(body)));
     else
       return expression_ptr(new 
          do_expression<false>(std::move(condition),std::move(body)));
   }
   break;
   
   case for_expr:
   { const for_node& f=*e.for_variant;
     type_expr in_type;
     expression_ptr in_expr = convert_expr(f.in_part,in_type);  // \&{in} part
     subscr_base::sub_type which; // the kind of aggregate iterated over
     layer bind(count_identifiers(f.id),nullptr);
      // for identifier(s) introduced in this loop
     
     { type_expr comp_type; const_type_p inx_type;
       which = subscr_base::index_kind(in_type,*(inx_type=&int_type),comp_type);
       if (which==subscr_base::not_so)
         // if not integer-indexable, try parameter-indexable
         which = subscr_base::index_kind(in_type,*(inx_type=&param_type),comp_type);
       if (which==subscr_base::not_so) // if its not that either, it is wrong
       { std::ostringstream o;
         o << "Cannot iterate over value of type " << in_type;
         throw expr_error(e,o.str());
       }
       type_list it_comps; // type of ``iterator'' value (pattern) named in the loop
       it_comps.push_front(std::move(comp_type));
       it_comps.push_front(type_expr(inx_type->copy()));
       type_expr it_type(std::move(it_comps));
         // build tuple type from index and component types
       if (not pattern_type(f.id).specialise(it_type))
         throw expr_error(e,"Improper structure of loop variable pattern");
       thread_bindings(f.id,it_type,bind,true); // force all identifiers constant
     }
     type_expr body_type;
     type_p btp; // will point to the place to record body type
     const conversion_record* conv=nullptr;
       // initialise in case we don't reach the assignment below
     if (type==void_type)
       btp=&type; // we can reuse this type; no risk of specialisation
     else if (type.specialise(row_of_type))
       btp=type.component_type;
     else if ((conv=row_coercion(type,body_type))!=nullptr)
       btp=&body_type;
     else throw type_error(e,row_of_type.copy(),std::move(type));
   
     expression_ptr body(convert_expr (f.body,*btp));
     if (type!=void_type and *btp==void_type and not is_empty(f.body))
       body.reset(new voiding(std::move(body)));
   expression_ptr loop;
     if (bind.empty()) // we must avoid having an empty |frame| produced at runtime
       
       { expression_ptr call; const source_location &loc = f.in_part.loc;
         switch(which)
         {
         case subscr_base::row_entry: call.reset(new 
             builtin_call(sizeof_row_builtin,std::move(in_expr),loc));
         break; case subscr_base::vector_entry: call.reset(new 
             builtin_call(sizeof_vector_builtin,std::move(in_expr),loc));
         break; case subscr_base::ratvec_entry: call.reset(new 
             builtin_call(sizeof_ratvec_builtin,std::move(in_expr),loc));
         break; case subscr_base::string_char: call.reset(new 
             builtin_call(sizeof_string_builtin,std::move(in_expr),loc));
         break; case subscr_base::matrix_column: call.reset(new 
             builtin_call(matrix_columns_builtin,std::move(in_expr),loc));
         break; case subscr_base::mod_poly_term: call.reset(new 
             builtin_call(sizeof_parampol_builtin,std::move(in_expr),loc));
         break; default: assert(false);
         }
       
         if (f.flags.test(1)) // whether reversed assembly of return value
           loop.reset(new  counted_for_expression<6>
             (-1,std::move(call),expression_ptr(nullptr),std::move(body)));
         else
           loop.reset(new  counted_for_expression<4>
             (-1,std::move(call),expression_ptr(nullptr),std::move(body)));
       }
     else loop.reset(make_for_loop
       (f.flags.to_ulong(),f.id,std::move(in_expr),std::move(body),which));
   return conv==nullptr ? std::move(loop)
     :  expression_ptr(new conversion(*conv,std::move(loop))) ;
   }
   
   case cfor_expr:
   { const cfor_node& c=*e.cfor_variant;
     expression_ptr count_expr = convert_expr(c.count,as_lvalue(int_type.copy()));
     expression_ptr bound_expr = is_empty(c.bound)
       ? expression_ptr(nullptr)
       : convert_expr(c.bound,as_lvalue(int_type.copy())) ;
   
     type_expr body_type;
     type_expr *btp=&body_type; // point to place to record body type
     const conversion_record* conv=nullptr;
     if (type==void_type)
       btp=&type; // we can reuse this type; no risk of specialisation
     else if (type.specialise(row_of_type))
       btp=type.component_type;
     else if ((conv=row_coercion(type,body_type))==nullptr)
       throw type_error(e,row_of_type.copy(),std::move(type));
   
     if (c.flags[2]) // case of absent loop variable
     { layer bind(0,nullptr);  // no local variables for loop, but allow |break|
       expression_ptr body(convert_expr (c.body,*btp));
       if (type!=void_type and *btp==void_type and not is_empty(c.body))
         body.reset(new voiding(std::move(body)));
     expression_ptr loop(make_counted_loop(c.flags.to_ulong(), 
         c.id,std::move(count_expr),std::move(bound_expr),std::move(body)));
       return conv==nullptr
              ? std::move(loop) 
              : expression_ptr(new conversion(*conv,std::move(loop)));
     }
     else // case of a present loop variable
     { layer bind(1,nullptr);
       bind.add(c.id,int_type.copy(),true); // add |id| as constant
       expression_ptr body(convert_expr (c.body,*btp));
       if (type!=void_type and *btp==void_type and not is_empty(c.body))
         body.reset(new voiding(std::move(body)));
     expression_ptr loop(make_counted_loop(c.flags.to_ulong(), 
         c.id,std::move(count_expr),std::move(bound_expr),std::move(body)));
       return conv==nullptr
              ? std::move(loop) 
              : expression_ptr(new conversion(*conv,std::move(loop)));
     }
   }
   
   case cast_expr:
   { cast_node& c=*e.cast_variant;
     if (type.specialise(c.type)) // see if we can do without conversion
       return convert_expr(c.exp,type); // in which case use now specialised |type|
     expression_ptr p = convert_expr(c.exp,c.type); // otherwise use |c.type|
     return conform_types(c.type,type,std::move(p),e);
   }
   
   case op_cast_expr:
   { const op_cast& c=e.op_cast_variant;
     const type_expr& ctype=c->type;
     std::ostringstream o;
     o << main_hash_table->name_of(c->oper) << '@' << ctype;
     const auto* entry = global_overload_table->entry(c->oper,ctype);
   
     if (entry!=nullptr) // something was found
     {
       expression_ptr p(new capture_expression(entry->value(),o.str()));
       const type_expr& res_t = entry->type().result_type;
       if (functype_specialise(type,ctype,res_t) or type==void_type)
         return p;
       throw type_error(e,type_expr(ctype.copy(),res_t.copy()),std::move(type));
     }
   // now we have no match from the overload table, try generic operations
     if (is_special_operator(c->oper))
       
       { if (c->oper==print_name())
         { if (functype_specialise(type,ctype,ctype))
           return expression_ptr(new  capture_expression (print_builtin,o.str()));
         }
         else if (c->oper==prints_name())
         { if (functype_specialise(type,ctype,void_type))
           return expression_ptr(new  capture_expression (prints_builtin,o.str()));
         }
         else if (c->oper==to_string_name())
         { if (functype_specialise(type,ctype,str_type))
           return expression_ptr(new  capture_expression (to_string_builtin,o.str()));
         }
         else if (c->oper==error_name())
         { if (functype_specialise(type,ctype,unknown_type))
           return expression_ptr(new  capture_expression (error_builtin,o.str()));
         }
         else if (c->oper==size_of_name())
           
           { if (ctype.kind==row_type)
             { if (functype_specialise(type,ctype,int_type))
             return expression_ptr(new 
                 capture_expression (sizeof_row_builtin,o.str()));
               throw type_error(e,ctype.copy(),std::move(type));
             }
             else if (is_pair_type(ctype))
             {
               type_expr& arg_tp0 = ctype.tupple->contents;
               type_expr& arg_tp1 = ctype.tupple->next->contents;
               if (arg_tp0.kind==row_type and *arg_tp0.component_type==arg_tp1)
               { if (functype_specialise(type,ctype,arg_tp0))
                   return expression_ptr(new 
                     capture_expression(suffix_elt_builtin,o.str()));
                 throw type_error(e,ctype.copy(),std::move(type));
               }
               if (arg_tp1.kind==row_type and *arg_tp1.component_type==arg_tp0)
               { if (functype_specialise(type,ctype,arg_tp1))
                 return expression_ptr(new 
                   capture_expression (prefix_elt_builtin,o.str()));
                 throw type_error(e,ctype.copy(),std::move(type));
               }
             }
           }
         else if (c->oper==concatenate_name())
           
           { if (ctype.kind==row_type and ctype.component_type->kind==row_type)
             { if (functype_specialise(type,ctype,*ctype.component_type))
             return expression_ptr(new 
                 capture_expression (join_rows_row_builtin,o.str()));
               throw type_error(e,ctype.copy(),std::move(type));
             }
             else if (is_pair_type(ctype))
             {
               type_expr& arg_tp0 = ctype.tupple->contents;
               type_expr& arg_tp1 = ctype.tupple->next->contents;
               if (arg_tp0.kind==row_type and arg_tp1==arg_tp0)
               { if (functype_specialise(type,ctype,arg_tp0))
                   return expression_ptr(new 
                     capture_expression (join_rows_builtin,o.str()));
                 throw type_error(e,ctype.copy(),std::move(type));
               }
             }
           }
       }
   throw program_error("No instance for "+o.str()+" found");
   }
   break;
   
   case ass_stat:
   if ( e.assign_variant->lhs.kind==0x1) // single identifier, do simple assign
   {
     id_type lhs=e.assign_variant->lhs.name;
     const_type_p id_t; size_t i,j; bool is_const;
     const bool is_local = (id_t=layer::lookup(lhs,i,j,is_const))!=nullptr;
     if (not is_local and (id_t=global_id_table->type_of(lhs,is_const))==nullptr)
       report_undefined(lhs,e,"assignment");
   
     if (is_const)
       report_constant_modified(lhs,e,"assignment");
   
   
     type_expr rhs_type = id_t->copy(); // provide a modifiable copy
     expression_ptr r(convert_expr(e.assign_variant->rhs,rhs_type));
     if (rhs_type!=*id_t)
       // assignment will specialise identifier, record to which type it does
     { if (is_local)
         layer::specialise(i,j,rhs_type);
         else
         global_id_table->specialise(lhs,rhs_type);
     }
     if (rhs_type==void_type and not is_empty(e.assign_variant->rhs))
       r.reset(new voiding(std::move(r)));
   
     expression_ptr assign = is_local
     ? expression_ptr(new local_assignment(lhs,i,j,std::move(r)))
   : expression_ptr(new global_assignment(lhs,std::move(r)));
     return conform_types(rhs_type,type,std::move(assign),e);
   }
   else 
        { const id_pat& pat=e.assign_variant->lhs;
          type_expr rhs_type;
          threader thr(e);
          thr.thread(pat,rhs_type);
          expression_ptr r = convert_expr(e.assign_variant->rhs,rhs_type);
          thr.refine();
          if (rhs_type==void_type and not is_empty(e.assign_variant->rhs))
            r.reset(new voiding(std::move(r)));
          expression_ptr m_ass (
            new  multiple_assignment
              (pat,std::move(r)
              ,thr.locs.undress(),thr.globs.undress(),std::move(thr.is_global)));
          return conform_types(rhs_type,type,std::move(m_ass),e);
        }
   
   
   case comp_ass_stat:
   { id_type aggr=e.comp_assign_variant->aggr;
     const expr& index=e.comp_assign_variant->index;
     const expr& rhs=e.comp_assign_variant->rhs;
   const_type_p aggr_t; size_t d,o; bool is_const;
     bool is_local = (aggr_t=layer::lookup(aggr,d,o,is_const))!=nullptr;
     if (not is_local and (aggr_t=global_id_table->type_of(aggr,is_const))==nullptr)
       report_undefined(aggr,e,"component assignment");
   
     if (is_const)
       report_constant_modified(aggr,e,"component assignment");
   
   
     type_expr ind_t;
     expression_ptr i = convert_expr(index,ind_t);
   type_expr comp_t;
     subscr_base::sub_type kind=subscr_base::index_kind(*aggr_t,ind_t,comp_t);
     if (not subscr_base::assignable(kind))
     { std::ostringstream o;
       o << "Cannot subscript value of type " << *aggr_t 
         << " with index of type " << ind_t << " in assignment";
       throw expr_error(e,o.str());
     }
     expression_ptr r = convert_expr(rhs,comp_t);
     if (aggr_t->kind==row_type)
       aggr_t->component_type->specialise(comp_t); // record type
     if (comp_t==void_type and not is_empty(rhs))
       r.reset(new voiding(std::move(r)));
     expression_ptr p;
     if (is_local)
       if (e.comp_assign_variant->reversed)
         p.reset(new local_component_assignment<true>
           (aggr,std::move(i),d,o,std::move(r),kind));
       else
         p.reset(new local_component_assignment<false>
           (aggr,std::move(i),d,o,std::move(r),kind));
     else
       if (e.comp_assign_variant->reversed)
         p.reset(new global_component_assignment<true>
           (aggr,std::move(i),std::move(r),kind));
       else
         p.reset(new global_component_assignment<false>
           (aggr,std::move(i),std::move(r),kind));
     return conform_types(comp_t,type,std::move(p),e);
   }
   
   case field_ass_stat:
   { id_type tupple=e.field_assign_variant->aggr;
     id_type sel =e.field_assign_variant->selector;
     const expr& rhs=e.field_assign_variant->rhs;
   const_type_p tuple_t; size_t d,o; bool is_const;
     bool is_local = (tuple_t=layer::lookup(tupple,d,o,is_const))!=nullptr;
     if (not is_local
         and (tuple_t=global_id_table->type_of(tupple,is_const))==nullptr)
       report_undefined(tupple,e,"field assignment");
   
     if (is_const)
       report_constant_modified(tupple,e,"field assignment");
   
    // Now get selector function from the overload table; ignore local bindings
     const projector_value* proj;
     { const auto* entry=global_overload_table->entry(sel,*tuple_t);
       if (entry==nullptr)
         throw expr_error (e,"Improper selection in field assignment");
       proj=dynamic_cast<const projector_value*>(entry->value().get());
       if (proj==nullptr)
         throw expr_error
           (e,"Selector in field assignment is not a projector function");
       assert(tuple_t->kind == tuple_type and
              proj->position < length(tuple_t->tupple));
     }
   
     type_p comp_loc;
       // we shall pass a modifiable reference to component type to |convert_expr|
     { // to get component type from list pointer we need to use a short loop
       auto p=tuple_t->tupple;
       for (auto count=proj->position; count-->0; )
         p=p->next.get();
       comp_loc=&p->contents;
     }
     expression_ptr r = convert_expr(rhs,*comp_loc);
     expression_ptr p;
     if (is_local)
       p.reset(new local_field_assignment(tupple,proj->position,d,o,std::move(r)));
     else
       p.reset(new global_field_assignment(tupple,proj->position,std::move(r)));
     return conform_types(*comp_loc,type,std::move(p),e);
   }
   case no_expr: assert(false);
  }
  return expression_ptr(nullptr); // keep compiler happy
}

void shell::evaluate (level l) const
{ throw runtime_error("I die"); } 

void breaker::evaluate (level l) const
{ throw loop_break(depth); }

void returner::evaluate (level l) const
{ exp->eval(); throw function_return(pop_value()); }

void tuple_expression::print(std::ostream& out) const
{ out << '(';
  for (auto it=component.begin(); it!=component.end(); ++it)
    out << (it==component.begin() ? "" : ",") << **it;
  out << ')';
}

void tuple_expression::evaluate(level l) const
{ switch(l)
  {
  case no_value:
    for (auto it=component.begin(); it!=component.end(); ++it)
      (*it)->void_eval();
    break;
  case single_value:
    { auto result = std::make_shared<tuple_value>(component.size());
      auto dst_it = result->val.begin();
      for (auto it=component.cbegin(); it!=component.cend(); ++it,++dst_it)
        { (*it)->eval(); *dst_it=pop_value(); }
      push_value(result);
    } break;
  case multi_value:
    { auto it=component.begin();
      try
      { for (; it!=component.end(); ++it)
          (*it)->eval();
      }
      catch (const loop_break&) // clean up execution stack locally
      { for (; it!=component.begin(); --it)
          execution_stack.pop_back();
        throw; // propagate the |break|
      }
      catch (const function_return&) // clean up execution stack locally
      { for (; it!=component.begin(); --it)
          execution_stack.pop_back();
        throw; // propagate the |break|
      }
    }
  } // |switch(l)|
}

void list_expression::print(std::ostream& out) const
{ out << '[';
  for (auto it=component.begin(); it!=component.end(); ++it)
    out << (it==component.begin() ? "" : ",") << **it;
  out << ']';
}

void list_expression::evaluate(level l) const
{ if (l==no_value)
    for (auto it=component.begin(); it!=component.end(); ++it)
      (*it)->void_eval();
  else
  { own_row result= std::make_shared<row_value>(0);
    result->val.reserve(component.size());
    for (auto it=component.begin(); it!=component.end(); ++it)
  {
      (*it)->eval();
      result->val.push_back(pop_value());
    }
    push_value(std::move(result)); // result will be shared from here on
  }
}

const char* identifier::name() const
{ return main_hash_table->name_of(code); }
void identifier::print(std::ostream& out) const
{ out<< name(); }

global_identifier::global_identifier(id_type id)
: identifier(id), address(global_id_table->address_of(id))
{}

void global_identifier::evaluate(level l) const
{ if (address->get()==nullptr)
  { std::ostringstream o;
    o << "Taking value of uninitialized variable '" << name() << '\'';
    throw runtime_error(o.str());
  }
  push_expanded(l,*address);
}

void local_identifier::evaluate(level l) const
{ push_expanded(l,frame::current->elem(depth,offset)); }

expression_ptr resolve_overload
  (const expr& e,
   type_expr& type,
   const overload_table::variant_list& variants)
{ const expr& args = e.call_variant->arg;
  type_expr apt; // modifiable temporary
  expression_ptr arg = convert_expr(args,apt);
    // get \foreign{a priori} type for argument
  const type_expr& a_priori_type = apt;
  id_type id =  e.call_variant->fun.identifier_variant;
  expression_ptr call(nullptr);
    // will be assigned in case of match requiring coercion
  const_type_p result_type = nullptr; // corresponding result type from table
  for (auto it=variants.begin(); it!=variants.end(); ++it)
  { const overload_data& v=*it;
    const auto& arg_type=v.type().arg_type;
    if (a_priori_type==arg_type) // exact match
    { 
      { if (arg_type==void_type and not is_empty(args))
          arg.reset(new voiding(std::move(arg)));
        std::ostringstream name;
        name << main_hash_table->name_of(id) << '@' << arg_type;
        call = v.value()->build_call(v.value(),name.str(),std::move(arg),e.loc);
      
        if (type==void_type and
            id==equals_name() and
            v.type().result_type!=void_type)
        { std::ostringstream o;
          o << "Use op equality operator '=' in void context; " 
            << "did you mean ':=' instead?\n  If you really want " 
            << "the result of '=' to be voided, use a cast to " 
            << v.type().result_type << '.';
          throw expr_error(e,o.str());
            // importantly this is not a |type_error|, which might be caught
        }
      }
      return conform_types(v.type().result_type,type,std::move(call),e);
      // exact match, return
    }
    else
    if ((is_close(a_priori_type,arg_type)&0x1)!=0)
    {
// inexact match, so tentatively convert again using |arg_type|, hoping coercion helps
      try
      { expression_ptr arg=convert_expr(args,as_lvalue(arg_type.copy()));
        // redo conversion
        
        { if (arg_type==void_type and not is_empty(args))
            arg.reset(new voiding(std::move(arg)));
          std::ostringstream name;
          name << main_hash_table->name_of(id) << '@' << arg_type;
          call = v.value()->build_call(v.value(),name.str(),std::move(arg),e.loc);
        
          if (type==void_type and
              id==equals_name() and
              v.type().result_type!=void_type)
          { std::ostringstream o;
            o << "Use op equality operator '=' in void context; " 
              << "did you mean ':=' instead?\n  If you really want " 
              << "the result of '=' to be voided, use a cast to " 
              << v.type().result_type << '.';
            throw expr_error(e,o.str());
              // importantly this is not a |type_error|, which might be caught
          }
        }
        result_type = &v.type().result_type;
        // just store the |call| and the its |result_type| now
      }
      catch (const type_error&) {}
       // if coercion fails ignore the match, but quit search anyway
      break;
    }
  }
  
  { std::ostringstream name;
    name << main_hash_table->name_of(id) << '@' << a_priori_type;
    if (id==size_of_name())
      
      { if (a_priori_type.kind==row_type)
        { expression_ptr call(new 
            builtin_call(sizeof_row_builtin,name.str(),std::move(arg),e.loc));
          return conform_types(int_type,type,std::move(call),e);
        }
        else if (is_pair_type(a_priori_type))
        {
          const type_expr& ap_tp0 = a_priori_type.tupple->contents;
          const type_expr& ap_tp1 = a_priori_type.tupple->next->contents;
          if (ap_tp0.kind==row_type and
              ap_tp0.component_type->specialise(ap_tp1)) // suffix case
          { expression_ptr call(new  builtin_call
              (suffix_elt_builtin,std::move(arg),e.loc));
            return conform_types(ap_tp0,type,std::move(call),e);
          }
          if (ap_tp1.kind==row_type and
              ap_tp1.component_type->specialise(ap_tp0)) // prefix case
          { expression_ptr call(new  builtin_call
              (prefix_elt_builtin,std::move(arg),e.loc));
            return conform_types(ap_tp1,type,std::move(call),e);
          }
        }
      }
    else if (id==concatenate_name())
      
      { if (a_priori_type.kind==row_type and
            a_priori_type.component_type->kind==row_type)
        { expression_ptr call(new 
            builtin_call(join_rows_row_builtin,std::move(arg),e.loc));
          return conform_types(*a_priori_type.component_type,type,std::move(call),e);
        }
        if (a_priori_type.kind==tuple_type and
          length(a_priori_type.tupple)==2 and
          a_priori_type.tupple->contents==a_priori_type.tupple->next->contents and
          a_priori_type.tupple->contents.kind==row_type)
        { expression_ptr call(new  builtin_call
              (join_rows_builtin,std::move(arg),e.loc));
          return conform_types
               (a_priori_type.tupple->contents,type,std::move(call),e);
        }
      }
    else // remaining cases always match
    { const bool needs_voiding = a_priori_type==void_type and not is_empty(args);
      if (id==print_name())
      { if (type!=void_type and not type.specialise(a_priori_type))
        {
          arg = convert_expr(args,type); // redo conversion, now in |type| context
          name.str(); name << "print@" << type; // correct |type| in |name|
        }
        return make_variadic_call
          (print_builtin,name.str(),std::move(arg),needs_voiding,e.loc);
      }
      else if(id==to_string_name()) // this always matches as well
      {
        expression_ptr call = make_variadic_call
          (to_string_builtin,name.str(),std::move(arg),needs_voiding,e.loc);
        return conform_types(str_type,type,std::move(call),e);
      }
      else if(id==prints_name()) // this always matches as well
      { expression_ptr call = make_variadic_call
         (prints_builtin,name.str(),std::move(arg),needs_voiding,e.loc);
        return conform_types(void_type,type,std::move(call),e);
        // check that |type==void_type|
      }
      else if(id==error_name()) // this always matches as well
        return make_variadic_call
          (error_builtin,name.str(),std::move(arg),needs_voiding,e.loc);
    }
  }
  if (result_type!=nullptr)
    // return inexact match if special operators did not do exact match
    return conform_types(*result_type,type,std::move(call),e);
  
  if (variants.size()==1)
    throw type_error(args,a_priori_type.copy(),variants[0].type().arg_type.copy());
  else
  { std::ostringstream o;
    o << "Failed to match '"
      << main_hash_table->name_of(id) 
      << "' with argument type "
      << a_priori_type;
    throw expr_error(e,o.str());
  }
}

void call_expression::print(std::ostream& out) const
{ if (dynamic_cast<const identifier*>(function.get())!=nullptr)
    out << *function;
  else out << '(' << *function << ')';
  if (dynamic_cast<const tuple_expression*>(argument.get())!=nullptr)
    out << *argument;
  else out << '(' << *argument << ')';
}

void overloaded_call::print(std::ostream& out) const
{ out << name;
  if (dynamic_cast<const tuple_expression*>(argument.get())!=nullptr)
     out << *argument;
  else out << '(' << *argument << ')';
}

template <bool variadic>
expression_ptr builtin_value<variadic>::build_call
    (const shared_function& owner,const std::string& name,
     expression_ptr&& arg, const source_location& loc) const
{ std::shared_ptr<const builtin_value<variadic> > f(owner,this);
  return expression_ptr(new 
    overloaded_builtin_call<variadic>(f,name,std::move(arg),loc));
}

template<>
void variadic_builtin_call::evaluate(level l) const
{ std::string arg_string;
  argument->eval(); ; // always evaluate to single value
  if (verbosity>0) // then record argument(s) as string
  { std::ostringstream o;
    o << *execution_stack.back();
    arg_string = o.str();
  }

  try
  { (*f_ptr)(l); } // call the built-in function
  
  catch (error_base& e)
  { extend_message(e,this,f,arg_string);
    throw;
  }
  catch (const std::exception& e)
  { runtime_error new_error(e.what());
    extend_message(new_error,this,f,arg_string);
    throw new_error;
  }
}

template<>
void builtin_call::evaluate(level l) const
{ std::string arg_string;
  if (verbosity==0)
    argument->multi_eval();
  else // record argument(s) as string
  { auto sp = execution_stack.size();
    argument->multi_eval(); // mark stack before evaluation
    std::ostringstream o;
    if (execution_stack.size()>sp+1) // multiple arguments
      for (o << '(';
           sp<execution_stack.size();
           o << (sp<execution_stack.size() ? ',' : ')')
          )
        o << *execution_stack[sp++];
    else if (execution_stack.size()==sp) o << "()"; // no arguments
    else
      o << *execution_stack.back(); // single argument case
    arg_string = o.str();
  }

  try
  { (*f_ptr)(l); } // call the built-in function
  
  catch (error_base& e)
  { extend_message(e,this,f,arg_string);
    throw;
  }
  catch (const std::exception& e)
  { runtime_error new_error(e.what());
    extend_message(new_error,this,f,arg_string);
    throw new_error;
  }
}

void call_expression::evaluate(level l) const
{ function->eval();
  auto f = std::dynamic_pointer_cast<const function_base>(pop_value());
  if (f.get()==nullptr)
    throw logic_error
      ("Non-function value found for function in a call expression");
  std::string arg_string;
  if (verbosity==0)
    argument->evaluate(f->argument_policy());
  else
  { auto sp = execution_stack.size();
    argument->evaluate(f->argument_policy());
    std::ostringstream o;
    if (execution_stack.size()>sp+1) // multiple arguments
      for (o << '(';
           sp<execution_stack.size();
           o << (sp<execution_stack.size() ? ',' : ')')
          )
        o << *execution_stack[sp++];
    else if (execution_stack.size()==sp) o << "()"; // no arguments
    else
      o << *execution_stack.back(); // single argument case
    arg_string = o.str();
  }

  try { f->apply(l); } // apply the function, handling |l| appropriately
  
  catch (error_base& e)
  { extend_message(e,this,f,arg_string);
    throw;
  }
  catch (const std::exception& e)
  { runtime_error new_error(e.what());
    extend_message(new_error,this,f,arg_string);
    throw new_error;
  }
}

inline
let_expression::let_expression 
  (const id_pat& v, expression_ptr&& ini, expression_ptr&& b)
: variable(copy_id_pat(v))
, initialiser(ini.release())
, body(b.release())
{}

void let_expression::print(std::ostream& out) const
{ out << " let " << variable << "=" << *initialiser << " in " << *body;
}

type_list pattern_list_types(const patlist& p)
{ dressed_type_list result;
  for (auto it = p.begin(); not p.at_end(it); ++it)
    result.push_back(pattern_type(*it));
  return result.undress();
}

type_expr pattern_type(const id_pat& pat)
{ return (pat.kind&0x2)==0
  ? unknown_type.copy()
  : type_expr(pattern_list_types(pat.sublist));
}

size_t count_identifiers(const id_pat& pat)
{ size_t result= pat.kind & 0x1; // 1 if |pat.name| is defined, 0 otherwise
  if ((pat.kind & 0x2)!=0) // then a list of subpatterns is present
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it)
      result+=count_identifiers(*it);
  return result;
}

void list_identifiers(const id_pat& pat, std::vector<id_type>& d)
{ if ((pat.kind & 0x1)!=0)
    d.push_back(pat.name);
  if ((pat.kind & 0x2)!=0) // then a list of subpatterns is present
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it)
      list_identifiers(*it,d);
}

void thread_bindings
(const id_pat& pat,const type_expr& type, layer& dst, bool is_const)
{ if ((pat.kind & 0x1)!=0)
    dst.add(pat.name,type.copy(),is_const or (pat.kind & 0x4)!=0);
  if ((pat.kind & 0x2)!=0)
  { assert(type.kind==tuple_type);
    wtl_const_iterator t_it(type.tupple);
    for (auto p_it=pat.sublist.begin(); not pat.sublist.at_end(p_it);
         ++p_it,++t_it)
      thread_bindings(*p_it,*t_it,dst,is_const);
  }
}

void thread_components
  (const id_pat& pat,const shared_value& val
  , std::back_insert_iterator<std::vector<shared_value> > dst)
{ if ((pat.kind & 0x1)!=0)
     *dst++ = val; // copy |shared_value| pointer |val|, creating sharing

  if ((pat.kind & 0x2)!=0)
  { const tuple_value* t=force<tuple_value>(val.get());
    assert(t->val.size()==length(pat.sublist));
    auto src = t->val.begin();
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it,++src)
      thread_components(*it,*src,dst);
  }
}

void let_expression::evaluate(level l) const
{ initialiser->eval(); // evaluate on stack as single value

  frame fr(variable); // save context, create new one for |f|
  fr.bind(pop_value()); // decompose arguments(s) and bind values in |fr|
  try { body->evaluate(l); }
    // call, passing evaluation level |l| to function body
  
  catch (error_base& e)
  { std::vector<id_type> names = fr.id_list();
    auto id_it = names.cbegin(); std::ostringstream o; o << "\n  [";
    for (auto it = frame::current->begin(); it!=frame::current->end();
         ++it,++id_it)
    { if (it!=frame::current->begin())
        o << ", ";
      o << main_hash_table->name_of(*id_it) << '=' << **it;
    }
    o << ']';
    e.message.append(o.str());
    throw;
  }
} 

inline
lambda_expression::lambda_expression 
  (const id_pat& p, expression_ptr&& b, const source_location& loc)
: p(std::make_shared<lambda_struct>(copy_id_pat(p),std::move(b),loc))
{}

std::ostream& operator<<(std::ostream& out, const lambda_struct& l)
{ if ((l.param.kind&0x1)!=0)
    if ((l.param.kind&0x4)!=0)
      out << "(!" << l.param << ')';
    else out << '(' << l.param << ')';
  else if ((l.param.kind&0x2)!=0 and not l.param.sublist.empty())
    out << l.param;
  else out << '@';
  return out << ": " << *l.body;
}
void lambda_expression::print(std::ostream& out) const { out << *p; }

void closure_value::print(std::ostream& out) const
{ out << "Function defined " << p->loc << std::endl << *p;
}

void closure_value::report_origin(std::ostream& o) const
  { o << "defined " << p->loc; }

void lambda_expression::evaluate(level l) const
{if (l!=no_value)
     push_value(std::make_shared<closure_value>(frame::current,p));
}

expression_ptr closure_value::build_call
    (const shared_function& owner,const std::string& name,
     expression_ptr&& arg, const source_location& loc) const
{ std::shared_ptr<const closure_value> f(owner,this);
  return expression_ptr(new  closure_call(f,name,std::move(arg),loc));
}

void closure_value::apply(expression_base::level l) const
{
  try
  { lambda_frame fr(p->param,context);
      // save context, create new one for |*this|
    if (fr.is_empty()) // we must test for functions without named arguments
    { execution_stack.pop_back();
      body.evaluate(l);
    } //  drop argument, evaluate avoiding |bind|
    else
    { fr.bind(pop_value()); // decompose arguments(s) and bind values in |fr|
      try { body.evaluate(l); }
        // call, passing evaluation level |l| to function body
      
      catch (error_base& e)
      { std::vector<id_type> names = fr.id_list();
        auto id_it = names.cbegin(); std::ostringstream o; o << "\n  [";
        for (auto it = frame::current->begin(); it!=frame::current->end();
             ++it,++id_it)
        { if (it!=frame::current->begin())
            o << ", ";
          o << main_hash_table->name_of(*id_it) << '=' << **it;
        }
        o << ']';
        e.message.append(o.str());
        throw;
      }
    }
  } // restore context upon destruction of |fr|
  
  catch (function_return& err) 
  { push_expanded(l,std::move(err.val)); }
}

void closure_call::evaluate(level l) const
{ argument->eval(); // evaluate arguments as a single value
  std::string arg_string;
  if (verbosity!=0) // then record argument(s) as string
  { std::ostringstream o;
    o << *execution_stack.back();
    arg_string = o.str();
  }
  try
  { lambda_frame fr(f->p->param,f->context);
      // save context, create new one for |f|

    if (fr.is_empty()) // we must test for functions without named arguments
      { execution_stack.pop_back(); f->body.evaluate(l); }
      //  avoid |bind|, evaluate
    else
    { fr.bind(pop_value()); // decompose arguments(s) and bind values in |fr|
      try { f->body.evaluate(l); }
      // call, passing evaluation level |l| to function body
      
      catch (error_base& e)
      { std::vector<id_type> names = fr.id_list();
        auto id_it = names.cbegin(); std::ostringstream o; o << "\n  [";
        for (auto it = frame::current->begin(); it!=frame::current->end();
             ++it,++id_it)
        { if (it!=frame::current->begin())
            o << ", ";
          o << main_hash_table->name_of(*id_it) << '=' << **it;
        }
        o << ']';
        e.message.append(o.str());
        throw;
      }
    }
  } // restore context upon destruction of |fr|
  
  catch (function_return& err) 
  { push_expanded(l,std::move(err.val)); }
  
  catch (error_base& e)
  { extend_message(e,this,f,arg_string);
    throw;
  }
  catch (const std::exception& e)
  { runtime_error new_error(e.what());
    extend_message(new_error,this,f,arg_string);
    throw new_error;
  }
}

void seq_expression::print(std::ostream& out) const
{ out << *first << ';' << *last; }
void next_expression::print(std::ostream& out) const
{ out << *first << " next " << *last; }

void seq_expression::evaluate(level l) const
{ first->void_eval(); last->evaluate(l); }

void next_expression::evaluate(level l) const
{ first->evaluate(l); last->void_eval(); }

void subscr_base::print(std::ostream& out,bool reversed) const
{ out << *array << (reversed ? "~[" : "[") << *index << ']';
}

template <bool reversed>
void matrix_subscription<reversed>::print(std::ostream& out) const
{ out  << *array << (reversed ? "~[" : "[");
  auto* p=dynamic_cast<const tuple_expression*>(index.get());
  if (p==nullptr) out << *index << ']';
  else
    out << *p->component[0] << ',' << *p->component[1] << ']';
}

void slice_base::print(std::ostream& out, unsigned flags) const
{  out << *array << ((flags&0x1)==0 ? "[" : "~[")
     << *lower << ((flags&0x2)==0 ? ":" : "~:")
     << *upper << ((flags&0x4)==0 ? "]" : "~]");
}

subscr_base::sub_type subscr_base::index_kind
  (const type_expr& aggr,
   const type_expr& index,
         type_expr& subscr)
{ if (aggr.kind==primitive_type)
    switch (aggr.prim)
    {  default:
    break; case vector_type:
      if (index==int_type and subscr.specialise(int_type))
        return vector_entry; 
    break; case rational_vector_type:
      if (index==int_type and subscr.specialise(rat_type))
        return ratvec_entry; 
    break; case string_type:
      if (index==int_type and subscr.specialise(str_type))
        return string_char; 
    break; case matrix_type:
      if (index==int_int_type and subscr.specialise(int_type))
        return matrix_entry;
      else if (index==int_type and subscr.specialise(vec_type))
        return matrix_column;
    break; case virtual_module_type:
      if (index==param_type and subscr.specialise(split_type))
        return mod_poly_term;
    }
  else if (aggr.kind==row_type and index==int_type and
           subscr.specialise(*aggr.component_type))
         return row_entry;
  return not_so;
}

subscr_base::sub_type subscr_base::slice_kind (const type_expr& aggr)
{ if (aggr.kind==primitive_type)
    switch (aggr.prim)
    {
    case vector_type: return vector_entry;
    case rational_vector_type: return ratvec_entry;
    case string_type: return string_char;
    case matrix_type: return matrix_column;
    default: return not_so;
    }
  else if (aggr.kind==row_type)
    return row_entry;
  else return not_so;
}

bool subscr_base::assignable(subscr_base::sub_type t)
{ switch (t)
  { case ratvec_entry: case string_char: case mod_poly_term:
    case not_so: return false;
    default: return true;
  }
}

inline std::string range_mess
  (int i,size_t n,const expression_base* e,const char* where)
{ std::ostringstream o;
  e->print(o << "index " << i << " out of range (0<= . <" << n
             << ") in " << where << ' ');
  return o.str();
}

template <bool reversed>
void row_subscription<reversed>::evaluate(level l) const
{ int i=(index->eval(),get<int_value>()->int_val());
  shared_row r=(array->eval(),get<row_value>());
  size_t n = r->val.size();
  if (reversed)
    i=n-1-i;
  if (static_cast<unsigned int>(i)>=n)
    throw runtime_error(range_mess(i,n,this,"subscription"));
  push_expanded(l,r->val[i]);
}

template <bool reversed>
void vector_subscription<reversed>::evaluate(level l) const
{ int i=(index->eval(),get<int_value>()->int_val());
  shared_vector v=(array->eval(),get<vector_value>());
  size_t n = v->val.size();
  if (reversed)
    i=n-1-i;
  if (static_cast<unsigned int>(i)>=n)
    throw runtime_error(range_mess(i,n,this,"subscription"));
  if (l!=no_value)
    push_value(std::make_shared<int_value>(v->val[i]));
}

template <bool reversed>
void ratvec_subscription<reversed>::evaluate(level l) const
{ int i=(index->eval(),get<int_value>()->int_val());
  shared_rational_vector v=(array->eval(),get<rational_vector_value>());
  size_t n = v->val.size();
  if (reversed)
    i=n-1-i;
  if (static_cast<unsigned int>(i)>=n)
    throw runtime_error(range_mess(i,n,this,"subscription"));
  if (l!=no_value)
    push_value(std::make_shared<rat_value>(Rational 
       (v->val.numerator()[i],v->val.denominator())));
}

template <bool reversed>
void string_subscription<reversed>::evaluate(level l) const
{ int i=(index->eval(),get<int_value>()->int_val());
  shared_string s=(array->eval(),get<string_value>());
  size_t n = s->val.size();
  if (reversed)
    i=n-1-i;
  if (static_cast<unsigned int>(i)>=n)
    throw runtime_error(range_mess(i,n,this,"subscription"));
  if (l!=no_value)
    push_value(std::make_shared<string_value>(s->val.substr(i,1)));
}

template <bool reversed>
void matrix_subscription<reversed>::evaluate(level l) const
{ index->multi_eval(); 
  int j=get<int_value>()->int_val();
  int i=get<int_value>()->int_val();
  shared_matrix m=(array->eval(),get<matrix_value>());
  size_t r = m->val.numRows();
  size_t c = m->val.numColumns();
  if (reversed)
  {  i=r-1-i; j=c-1-j; }
  if (static_cast<unsigned int>(i)>=r)
    throw runtime_error
     ("initial "+range_mess(i,r,this,"matrix subscription"));
  if (static_cast<unsigned int>(j)>=c)
    throw runtime_error
     ("final "+range_mess(j,c,this,"matrix subscription"));
  if (l!=no_value)
    push_value(std::make_shared<int_value>(m->val(i,j)));
}

template <bool reversed>
void matrix_get_column<reversed>::evaluate(level l) const
{ int j=(index->eval(),get<int_value>()->int_val());
  shared_matrix m=(array->eval(),get<matrix_value>());
  size_t c = m->val.numColumns();
  if (reversed)
    j=c-1-j;
  if (static_cast<unsigned int>(j)>=c)
    throw runtime_error(range_mess(j,c,this,"matrix column selection"));
  if (l!=no_value)
    push_value(std::make_shared<vector_value>(m->val.column(j)));
}

void slice_range_error
  (int lwb, int upb,int n,unsigned flags,const expression_base* e)
{ std::ostringstream o;
  if (upb>n)
    if (lwb<0)
      o << "both bounds " << lwb << ':' << upb;
    else
      o << "upper bound " << upb;
  else
    o << "lower bound " << lwb;
  o << " out of range (should be ";
  if (lwb<0)
    o << ">=0" << (upb>n ? " respectively " : ")");
  if (upb>n)
    o << "<=" << n << ')';
  e->print(o << " in slice ");
  throw runtime_error(o.str());
}

template <unsigned flags>
void row_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->int_val());
  int lwb=(lower->eval(),get<int_value>()->int_val());
  shared_row arr=(array->eval(),get<row_value>());
  const auto& r = arr->val;
  int n = r.size();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
  { push_value(std::make_shared<row_value>(0)); return; }
  push_value( (flags&0x1)==0
  ? std::make_shared<row_value>(r.begin()+lwb,r.begin()+upb)
  : std::make_shared<row_value>(r.rbegin()+lwb,r.rbegin()+upb)
    );
}

template <unsigned flags>
void vector_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->int_val());
  int lwb=(lower->eval(),get<int_value>()->int_val());
  shared_vector arr=(array->eval(),get<vector_value>());
  const auto& r = arr->val;
  int n = r.size();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
  { push_value(std::make_shared<vector_value>(int_Vector(0))); return; }
  push_value( (flags&0x1)==0
  ? std::make_shared<vector_value>(r.begin()+lwb,r.begin()+upb)
  : std::make_shared<vector_value>(r.rbegin()+lwb,r.rbegin()+upb)
    );
}

template <unsigned flags>
void ratvec_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->int_val());
  int lwb=(lower->eval(),get<int_value>()->int_val());
  shared_rational_vector arr=(array->eval(),get<rational_vector_value>());
  const auto& r = arr->val.numerator();
  int n = r.size();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
  { push_value(std::make_shared<rational_vector_value>(RatWeight(0)));
      return; }
  const int d=arr->val.denominator();
  push_value( (flags&0x1)==0
  ? std::make_shared<rational_vector_value>(r.begin()+lwb,r.begin()+upb,d)
  : std::make_shared<rational_vector_value>(r.rbegin()+lwb,r.rbegin()+upb,d)
    );
}

template <unsigned flags>
void string_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->int_val());
  int lwb=(lower->eval(),get<int_value>()->int_val());
  shared_string arr=(array->eval(),get<string_value>());
  const auto& r = arr->val;
  int n = r.size();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
  { push_value(std::make_shared<string_value>(std::string())); return; }
  push_value
    ((flags&0x1)==0
  ? std::make_shared<string_value>(std::string(&r[lwb],&r[upb]))
  : std::make_shared<string_value>(
       std::string(r.rbegin()+lwb,r.rbegin()+upb)));
}

template <unsigned flags>
void matrix_slice<flags>::evaluate(level l) const
{ int upb=(upper->eval(),get<int_value>()->int_val());
  int lwb=(lower->eval(),get<int_value>()->int_val());
  shared_matrix mat=(array->eval(),get<matrix_value>());
  const auto& m = mat->val;
  int n = m.numColumns();
  if ((flags&0x2)!=0)
    lwb = n - lwb;
  if ((flags&0x4)!=0)
    upb = n - upb;
  if (lwb<0 or upb>n)
    slice_range_error(lwb,upb,n,flags,this);
  if (lwb>=upb)
{ push_value(std::make_shared<matrix_value>(int_Matrix(m.numRows(),0)));
      return; }
  own_matrix result =
    std::make_shared<matrix_value>(int_Matrix(m.numRows(),upb-lwb));
  auto& r = result->val;
  for (unsigned int j=0; lwb<upb; ++j)
    r.set_column(j,m.column((flags&0x1)==0 ? lwb++ : n - ++lwb));
  push_value(std::move(result));
}

void projector_value::print(std::ostream& out) const
  { out << "{."<< main_hash_table->name_of(id) << ": "
                 << type << '.' << position << '}'; }
expression_base::level projector_value::argument_policy() const
  { return expression_base::single_value; }
void projector_value::report_origin(std::ostream& o) const
 { o << "projector defined " << loc; }

void projector_value::apply(expression_base::level l) const
{ push_expanded(l,get<tuple_value>()->val[position]); }

expression_ptr projector_value::build_call
    (const shared_function&,const std::string& name,
     expression_ptr&& arg, const source_location& loc) const
{ return expression_ptr(new projector_call(*this,name,std::move(arg),loc));
}

void projector_call::evaluate(level l) const
{
  argument->eval(); // evaluate arguments as a single value
  push_expanded(l,get<tuple_value>()->val[position]);
}

void projector_call::print(std::ostream& out) const
{ out << *argument << '.' << main_hash_table->name_of(id); }

void injector_value::print(std::ostream& out) const
  { out << "{."<< main_hash_table->name_of(id) << ": "
                 << position << type << '}'; }
expression_base::level injector_value::argument_policy() const
  { return expression_base::single_value; }
void injector_value::report_origin(std::ostream& o) const
 { o << "injector defined " << loc; }

void injector_value::apply(expression_base::level l) const
{ shared_value component = pop_value();
  push_value(std::make_shared<union_value>(position,std::move(component),id));
}

expression_ptr injector_value::build_call
    (const shared_function&,const std::string& name,
     expression_ptr&& arg, const source_location& loc) const
{ return expression_ptr(new injector_call(*this,name,std::move(arg),loc));
}

void injector_call::evaluate(level l) const
{
  argument->eval(); // evaluate arguments as a single value
  push_value(std::make_shared<union_value>(position,pop_value(),id));
}

void injector_call::print(std::ostream& out) const
{ out << *argument << '.' << main_hash_table->name_of(id); }

void conditional_expression::print(std::ostream& out) const
{ out << " if "; const conditional_expression* cur=this;
  do
  { out << *cur->condition << " then " << *cur->then_branch;
    auto* p =
      dynamic_cast<const conditional_expression*>(cur->else_branch.get());
    if (p==nullptr)
      break;
    out << " elif "; cur=p;
  }
  while(true);
  out << " else " << *cur->else_branch << " fi ";
}

void conditional_expression::evaluate(level l) const
{ condition->eval();
  if (get<bool_value>()->val)
    then_branch->evaluate(l);
   else else_branch->evaluate(l);
}

void int_case_expression::print(std::ostream& out) const
{ auto it = branches.cbegin();
  assert(it!=branches.cend());
  out << " case " << *condition << " in " << **it;
  while (++it!=branches.cend())
    out << ", " << **it;
  out << " esac ";
}

void int_case_expression::evaluate(level l) const
{ condition->eval();
  int i = get<int_value>()->int_val();
  if (static_cast<unsigned>(i)>=branches.size())
    throw runtime_error(range_mess(i,branches.size(),this,"case expression"));
  branches[i]->evaluate(l);
}

void union_case_expression::print(std::ostream& out) const
{ auto it = branches.cbegin();
  assert(it!=branches.cend());
  out << " case " << *condition << " in " << **it;
  while (++it!=branches.cend())
    out << " | " << **it;
  out << " esac ";
}

void discrimination_expression::print(std::ostream& out) const
{ auto it = branches.cbegin();
  assert(it!=branches.cend());
  out << " case " << *condition;
  do out << " | (" << it->first << "):" << *it->second;
  while (++it!=branches.cend());
  out << " esac ";
}

void union_case_expression::evaluate(level l) const
{ condition->eval();
  shared_union discriminant = get<union_value>();
  const auto& branch = branches[discriminant->variant()]; // make selection
  branch->evaluate(l); // evaluate function value
  shared_function f = get<function_base>();
  push_expanded(f->argument_policy(),discriminant->contents());
  f->apply(l);
}



void discrimination_expression::evaluate(level l) const
{ condition->eval();
  shared_union discriminant = get<union_value>();
  const auto& branch = branches[discriminant->variant()]; // make selection
  if (branch.first.kind==0x0)
    branch.second->evaluate(l); // avoid creating an empty |frame|
  else
  { frame fr(branch.first);
    fr.bind(discriminant->contents());
    branch.second->evaluate(l);
  }
}

template<unsigned flags>
void while_expression<flags>::print(std::ostream& out) const
{ out << " while " << *body << (out_reversed(flags) ? " ~od " : " od "); }

template <bool negated>
  void do_expression<negated>::print(std::ostream& out) const
{ out << *condition << (negated ? " ~do " : " do ") << *body; }
void forever_expression::print(std::ostream& out) const 
{ out << " do " << *body; }
void dont_expression::print(std::ostream& out) const 
{ out << " dont "; }

template <>
  void do_expression<false>::evaluate(level l) const
{ condition->eval();
  if (get<bool_value>()->val)
    { body->evaluate(l); while_condition_result=true; }
  else while_condition_result=false;
}
template <>
  void do_expression<true>::evaluate(level l) const
{ condition->eval();
  if (get<bool_value>()->val)
     while_condition_result=false;
  else { body->evaluate(l); while_condition_result=true; }
}

void forever_expression::evaluate(level l) const
{ body->evaluate(l); while_condition_result=true; }
void dont_expression::evaluate(level l) const
{ while_condition_result=false; }

template<unsigned flags>
void while_expression<flags>::evaluate(level l) const
{ if (l==no_value)
  { try
    { do
      { body->void_eval();
        check_interrupt();
      }
      while (while_condition_result);
    }
    catch (loop_break& err) 
    { if (err.depth-- > 0)
          throw;
    }
  }
  else if (yields_count(flags))
  { int count=0;
    try
    { while(body->void_eval(), while_condition_result)
      { ++count;
        check_interrupt();
      }
    }
    catch (loop_break& err) 
    { if (err.depth-- > 0)
          throw;
    }
    push_value(std::make_shared<int_value>(count));
  }
  else
  
  { containers::simple_list<shared_value> result;
    size_t s=0;
    try
    { while (body->eval(),while_condition_result)
      { check_interrupt();
        result.push_front(pop_value());
        ++s;
      }
    }
    catch (loop_break& err) 
    { if (err.depth-- > 0)
          throw;
    }
    own_row r = std::make_shared<row_value>(s);
    if (out_forward(flags)) // forward accumulating while loop
    { auto dst = r->val.rbegin();
        // use reverse iterator, since we reverse accumulated
      for (auto it = result.wbegin(); not it.at_end(); ++it,++dst)
        *dst = std::move(*it);
    }
    else // reverse accumulating while loop
    { auto dst = r->val.begin(); // use ordinary iterator
      for (auto it = result.wbegin(); not it.at_end(); ++it,++dst)
        *dst = std::move(*it);
    }
    push_value(std::move(r));
  }
}

template <unsigned flags, subscr_base::sub_type kind>
for_expression<flags,kind>::for_expression
 (const id_pat& p, expression_ptr&& i, expression_ptr&& b)
   : pattern(copy_id_pat(p)), in_part(i.release()), body(b.release())
  {}

void print_body(std::ostream& out,const expression_ptr& body,unsigned flags)
{ out << (in_reversed(flags) ? " ~do " : " do ")  << *body
        << (out_reversed(flags) ? " ~od " : " od ");
}

template <unsigned flags, subscr_base::sub_type kind>
void for_expression<flags,kind>::print(std::ostream& out) const
{ out << " for " << *std::next(pattern.sublist.begin());
    if (pattern.sublist.front().kind==0x1)
      out << '@' << pattern.sublist.front();
  print_body(out << " in " << *in_part,body,flags);
}

template <unsigned flags, subscr_base::sub_type kind>
void for_expression<flags,kind>::evaluate(level l) const
{ in_part->eval();
  own_tuple loop_var = std::make_shared<tuple_value>(2);
       // safe to re-use among iterations
  own_row result;
  std::vector<shared_value>::iterator dst;
  // set to |result->val.begin()| or |result->val.end()|
  try
  { switch (kind)
    {
      
      case subscr_base::row_entry:
        { shared_row in_val = get<row_value>();
          size_t n=in_val->val.size();
          
          size_t i= in_forward(flags) ? 0 : n;
          if (l!=no_value)
          { result = std::make_shared<row_value>(n);
            dst = out_forward(flags) ? result->val.begin() : result->val.end();
          }
          while (i!=(in_forward(flags) ? n : 0))
          { loop_var->val[1]=in_val->val[in_forward(flags) ? i : i-1];
              // move index into |loop_var| pair
            
            { loop_var->val[0] = std::make_shared<int_value>(in_forward(flags) ? i++ : --i);
                // create a fresh index each time
              frame loop_frame (pattern);
              loop_frame.bind(loop_var);
              if (l==no_value)
                body->void_eval();
              else
              { body->eval();
                 *(out_forward(flags) ? dst++ : --dst) = pop_value();
              }
            } 
          }
        }
        break;
      case subscr_base::vector_entry:
        { shared_vector in_val = get<vector_value>();
          size_t n=in_val->val.size();
          
          size_t i= in_forward(flags) ? 0 : n;
          if (l!=no_value)
          { result = std::make_shared<row_value>(n);
            dst = out_forward(flags) ? result->val.begin() : result->val.end();
          }
          while (i!=(in_forward(flags) ? n : 0))
          { loop_var->val[1] = std::make_shared<int_value>
              (in_val->val[in_forward(flags) ? i : i-1]);
            
            { loop_var->val[0] = std::make_shared<int_value>(in_forward(flags) ? i++ : --i);
                // create a fresh index each time
              frame loop_frame (pattern);
              loop_frame.bind(loop_var);
              if (l==no_value)
                body->void_eval();
              else
              { body->eval();
                 *(out_forward(flags) ? dst++ : --dst) = pop_value();
              }
            } 
          }
        }
        break;
      case subscr_base::ratvec_entry:
        { shared_rational_vector in_val = get<rational_vector_value>();
          size_t n=in_val->val.size();
          
          size_t i= in_forward(flags) ? 0 : n;
          if (l!=no_value)
          { result = std::make_shared<row_value>(n);
            dst = out_forward(flags) ? result->val.begin() : result->val.end();
          }
          while (i!=(in_forward(flags) ? n : 0))
          { loop_var->val[1] = std::make_shared<rat_value> 
            (Rational
              (in_val->val.numerator()[in_forward(flags) ? i : i-1]
              ,in_val->val.denominator()));
            
            { loop_var->val[0] = std::make_shared<int_value>(in_forward(flags) ? i++ : --i);
                // create a fresh index each time
              frame loop_frame (pattern);
              loop_frame.bind(loop_var);
              if (l==no_value)
                body->void_eval();
              else
              { body->eval();
                 *(out_forward(flags) ? dst++ : --dst) = pop_value();
              }
            } 
          }
        }
        break;
      case subscr_base::string_char:
        { shared_string in_val = get<string_value>();
          size_t n=in_val->val.size();
          
          size_t i= in_forward(flags) ? 0 : n;
          if (l!=no_value)
          { result = std::make_shared<row_value>(n);
            dst = out_forward(flags) ? result->val.begin() : result->val.end();
          }
          while (i!=(in_forward(flags) ? n : 0))
          { loop_var->val[1] = std::make_shared<string_value>
                  (in_val->val.substr(in_forward(flags) ? i : i-1,1));
            
            { loop_var->val[0] = std::make_shared<int_value>(in_forward(flags) ? i++ : --i);
                // create a fresh index each time
              frame loop_frame (pattern);
              loop_frame.bind(loop_var);
              if (l==no_value)
                body->void_eval();
              else
              { body->eval();
                 *(out_forward(flags) ? dst++ : --dst) = pop_value();
              }
            } 
          }
        }
        break;
      
      case subscr_base::matrix_column:
        { shared_matrix in_val = get<matrix_value>();
          size_t n=in_val->val.numColumns();
          
          size_t i= in_forward(flags) ? 0 : n;
          if (l!=no_value)
          { result = std::make_shared<row_value>(n);
            dst = out_forward(flags) ? result->val.begin() : result->val.end();
          }
          while (i!=(in_forward(flags) ? n : 0))
          { loop_var->val[1] = std::make_shared<vector_value>
              (in_val->val.column(in_forward(flags) ? i : i-1));
            
            { loop_var->val[0] = std::make_shared<int_value>(in_forward(flags) ? i++ : --i);
                // create a fresh index each time
              frame loop_frame (pattern);
              loop_frame.bind(loop_var);
              if (l==no_value)
                body->void_eval();
              else
              { body->eval();
                 *(out_forward(flags) ? dst++ : --dst) = pop_value();
              }
            } 
          }
        }
        break;
      case subscr_base::mod_poly_term:
        
        { shared_virtual_module pol_val = get<virtual_module_value>();
          size_t n=pol_val->val.size();
          if (l!=no_value)
          { result = std::make_shared<row_value>(n);
            dst = out_forward(flags) ? result->val.begin() : result->val.end();
          }
          if (in_forward(flags))
            for (auto it=pol_val->val.cbegin(); it!=pol_val->val.cend(); ++it)
              
              { loop_var->val[0] =
                  std::make_shared<module_parameter_value>(pol_val->rf,it->first);
                loop_var->val[1] = std::make_shared<split_int_value>(it->second);
                frame loop_frame(pattern);
                loop_frame.bind(loop_var);
                if (l==no_value)
                  body->void_eval();
                else
                { body->eval();
                  *(out_forward(flags) ? dst++ : --dst) = pop_value();
                }
              } 
          else
            for (auto it=pol_val->val.crbegin(); it!=pol_val->val.crend(); ++it)
              
              { loop_var->val[0] =
                  std::make_shared<module_parameter_value>(pol_val->rf,it->first);
                loop_var->val[1] = std::make_shared<split_int_value>(it->second);
                frame loop_frame(pattern);
                loop_frame.bind(loop_var);
                if (l==no_value)
                  body->void_eval();
                else
                { body->eval();
                  *(out_forward(flags) ? dst++ : --dst) = pop_value();
                }
              } 
        }
        break;
      case subscr_base::matrix_entry:; // excluded in type analysis
      case subscr_base::not_so: assert(false);
    }
  }
  catch (loop_break& err)
  { if (err.depth-- > 0)
      throw;
    if (l!=no_value)
    { if (out_reversed(flags))
        // doing |break| in reverse-gathering loop requires a shift
        dst=std::move(dst,result->val.end(),result->val.begin());
          // left-align |result|
      result->val.resize(dst-result->val.begin());
      // resize needed in all cases
    }
  }

  if (l!=no_value)
    push_value(std::move(result));
}

template <unsigned flags>
void counted_for_expression<flags>::print(std::ostream& out) const
{ if (has_frame(flags))
    out << " for " << main_hash_table->name_of(id) << ": " << *count;
  else out << " for : " << *count;  // omit nonexistent identifier
  if (bound.get()!=nullptr)
    out << " from " << *bound;
  print_body(out,body,flags);
}

template <unsigned flags>
void counted_for_expression<flags>::evaluate(level l) const
{ int c=(count->eval(),get<int_value>()->int_val());
  if (c<0)
    c=0; // no negative size result

  if (has_frame(flags)) // then loop uses index
  { int b=(bound.get()==nullptr ? 0 : (bound->eval(),get<int_value>()->int_val()));
    id_pat pattern(id);
    if (l==no_value)
      
      { c+=b; // set to upper bound, exclusive
        try
        { if (in_forward(flags)) // increasing loop
            while (b<c)
            { frame fr(pattern);
              fr.bind(std::make_shared<int_value>(b++));
              body->void_eval();
            }
          else if (b!=0)
            while (c-->b)
            { frame fr(pattern);
              fr.bind(std::make_shared<int_value>(c));
              body->void_eval();
            }
          else // same with |b==0|, but this is marginally faster
             while (c-->0)
            { frame fr(pattern);
              fr.bind(std::make_shared<int_value>(c));
              body->void_eval();
            }
        }
        catch (loop_break& err) 
        { if (err.depth-- > 0)
              throw;
        }
      }
    else
      
      { own_row result = std::make_shared<row_value>(c);
        c+=b; // set to upper bound, exclusive
        auto dst = out_forward(flags) ? result->val.begin() : result->val.end();
        try
        { if (in_forward(flags)) // increasing loop
            while (b<c)
            { frame fr(pattern);
              fr.bind(std::make_shared<int_value>(b++));
              body->eval();
              *(out_forward(flags)? dst++:--dst) = pop_value();
            }
          else if (b!=0)
            while (c-->b)
            { frame fr(pattern);
              fr.bind(std::make_shared<int_value>(c));
              body->eval();
              *(out_forward(flags)? dst++:--dst) = pop_value();
            }
          else // same with |b==0|, but this is marginally faster
            while (c-->0)
            { frame fr(pattern);
              fr.bind(std::make_shared<int_value>(c));
              body->eval();
              *(out_forward(flags)? dst++:--dst) = pop_value();
           }
        }
        catch (loop_break& err)
        { if (err.depth-- > 0)
            throw;
          if (out_reversed(flags))
            dst=std::move(dst,result->val.end(),result->val.begin());
            // after break, left-align |result|
          result->val.resize(dst-result->val.begin());
        }
        push_value(std::move(result));
      }
  }
  else if (l==no_value) // counted loop without index and no value
  { try { while (c-->0)
      body->void_eval();
    }
    catch (loop_break& err) 
    { if (err.depth-- > 0)
          throw;
    }
  }
  else // counted loop without index producing a value
  { own_row result = std::make_shared<row_value>(c);
    auto dst = out_forward(flags) ? result->val.begin() : result->val.end();
    try {
      while (c-->0)
      { body->eval();
        *(out_forward(flags)? dst++:--dst) = pop_value();
      }
    }
    catch (loop_break& err)
    { if (err.depth-- > 0)
        throw;
      if (out_reversed(flags))
        dst=std::move(dst,result->val.end(),result->val.begin());
        // after break, left-align |result|
      result->val.resize(dst-result->val.begin());
    }
    push_value(std::move(result));
  }

}

void assignment_expr::print(std::ostream& out) const
{ out << main_hash_table->name_of(lhs) << ":=" << *rhs; }

global_assignment::global_assignment(id_type l,expression_ptr&& r)
: assignment_expr(l,std::move(r)), address(global_id_table->address_of(l)) {}

void global_assignment::evaluate(level l) const
{ rhs->eval();
  *address = pop_value();
  push_expanded(l,*address);
}

local_assignment::local_assignment
 (id_type l, size_t i,size_t j, expression_ptr&& r)
: assignment_expr(l,std::move(r)), depth(i), offset(j) {}

void local_assignment::evaluate(level l) const
{ rhs->eval();
  shared_value& dest =  frame::current->elem(depth,offset);
  dest= pop_value();
  push_expanded(l,dest);
}

multiple_assignment::multiple_assignment 
    (const id_pat& lhs,expression_ptr&& r
    ,loc_list&& ll, glob_list&& gl, BitMap&& bm)
: lhs(copy_id_pat(lhs)), rhs(std::move(r))
, locals(std::move(ll)),globals(std::move(gl)), is_global(std::move(bm)){}

void multiple_assignment::print(std::ostream& out) const
{ out << "set " << lhs << ":=" << *rhs; }

void multiple_assignment::evaluate (level l) const
{ rhs->eval();
  shared_value v = pop_value();
  dest_iterator dst(locals,globals,is_global);
  thread_assign(lhs,v,dst);
  push_expanded(l,v);
}

void threader::thread(const id_pat& pat,type_expr& type)
{ if ((pat.kind&0x4)!=0)
    
    { std::ostringstream o;
      o << "Cannot constant-qualify '!' identifier '"
        << main_hash_table->name_of(pat.name)  << "' in multi-assignment";
      throw expr_error(e,o.str());
    }
  if ((pat.kind&0x2)!=0) // first treat any sublist
  { type.specialise(unknown_tuple(length(pat.sublist)));
    assert(type.kind==tuple_type); // this should succeed
    wtl_iterator t_it(type.tupple);
    for (auto it=pat.sublist.begin(); not pat.sublist.at_end(it); ++it,++t_it)
      thread(*it,*t_it);
  }
  if ((pat.kind&0x1)!=0)
  { id_type id = pat.name;
    const_type_p id_t; // will point to type of local or global |id|
    
    for (auto it=assoc.wcbegin(); not it.at_end(); ++it)
      if (it->first==id)
       { std::ostringstream o;
         o << "Multiple assignments to same identifier '"
           << main_hash_table->name_of(pat.name)  << "' in multi-assignment";
         throw expr_error(e,o.str());
       }
    size_t i,j; bool is_const;
    const bool is_local = (id_t=layer::lookup(id,i,j,is_const))!=nullptr;
    if (not is_local and (id_t = global_id_table->type_of(id,is_const))==nullptr)
      report_undefined(id,e,"multiple assignment");
    if (is_const)
      report_constant_modified(id,e,"multiple assignment");
    is_global.extend_capacity(not is_local);

    if (not type.specialise(*id_t)) // incorporate found type into |type|
      
      { std::ostringstream o;
        o << "Incompatible type for '" << main_hash_table->name_of(id)
        << "' in multi-assignment: type " << *id_t
        << " does no match pattern " << type;
        throw expr_error(e,o.str());
      }
    assoc.push_back(std::make_pair(id,&type));
      // record pointer to |type| for later refinement of |id|
    if (is_local)
      locs.push_back(multiple_assignment::local_dest{i,j});
    else
      globs.push_back(global_id_table->address_of(id));
  }
}

void threader::refine() const
{ unsigned long n=0;
  auto loc_it = locs.cbegin();
  for (auto it = assoc.cbegin(); not assoc.at_end(it); ++it)
    if (is_global.isMember(n++))
      global_id_table->specialise(it->first,*it->second);
    else
    {
      layer::specialise(loc_it->depth,loc_it->offset,*it->second);
      ++loc_it;
    }
}

template <bool reversed>
void component_assignment<reversed>::print (std::ostream& out) const
{ out << main_hash_table->name_of(lhs) << (reversed ? "~[" : "[")
        << *index << "]:=" << *rhs;
}

void field_assignment::print (std::ostream& out) const
{ out << main_hash_table->name_of(lhs) << '.' << this->position << ":="
        << *rhs;
}

template <bool reversed>
global_component_assignment<reversed>::global_component_assignment 
  (id_type a,expression_ptr&& i,expression_ptr&& r, subscr_base::sub_type k)
: base(a,std::move(i),std::move(r))
, kind(k),address(global_id_table->address_of(a)) {}

global_field_assignment::global_field_assignment 
  (id_type a, unsigned pos,expression_ptr&& r)
: field_assignment(a,pos,std::move(r))
, address(global_id_table->address_of(a)) {}

template <bool reversed>
void global_component_assignment<reversed>::evaluate(expression_base::level l)
  const
{ if (address->get()==nullptr)
  { std::ostringstream o;
    o << "Assigning to component of uninitialized variable " 
      << main_hash_table->name_of(this->lhs);
    throw runtime_error(o.str());
  }
  base::assign(l,*address,kind);
}

void global_field_assignment::evaluate(expression_base::level l) const
{ if (address->get()==nullptr)
  { std::ostringstream o;
    o << "Assigning to field of uninitialized variable " 
      << main_hash_table->name_of(this->lhs);
    throw runtime_error(o.str());
  }
  assign(l,*address);
}

template <bool reversed>
void component_assignment<reversed>::assign
  (level lev,shared_value& aggregate, subscr_base::sub_type kind) const
{ rhs->eval();
  value_base* loc=uniquify(aggregate);
    // raw pointer to modifiable value from shared pointer
  switch (kind)
  { case subscr_base::row_entry:
  
  { unsigned int i=(index->eval(),get<int_value>()->int_val());
    std::vector<shared_value>& a=force<row_value>(loc)->val;
    size_t n=a.size();
    if (i>=n)
      throw runtime_error(range_mess(i,a.size(),this,"component assignment"));
    auto& ai = a[reversed ? n-1-i : i];
    ai = pop_value(); // assign non-expanded value
    push_expanded(lev,ai); // return value may need expansion, or be omitted
  }
  break;
    case subscr_base::vector_entry:
  
  { unsigned int i=(index->eval(),get<int_value>()->int_val());
    std::vector<int>& v=force<vector_value>(loc)->val;
    size_t n=v.size();
    if (i>=n)
      throw runtime_error(range_mess(i,v.size(),this,"component assignment"));
    v[reversed ? n-1-i : i]=
      force<int_value>(execution_stack.back().get())->int_val();
      // assign |int| from un-popped top
    if (lev==no_value)
      execution_stack.pop_back(); // pop it anyway if result not needed
  }
  break;
    case subscr_base::matrix_entry:
  
  { index->multi_eval();
    unsigned int j=get<int_value>()->int_val();
    unsigned int i=get<int_value>()->int_val();
  
    int_Matrix& m=force<matrix_value>(loc)->val;
    size_t k=m.numRows(),l=m.numColumns();
    if (i>=k)
      throw runtime_error(range_mess(i,m.numRows(),this,"matrix entry assignment"));
    if (j>=l)
      throw runtime_error(
        range_mess(j,m.numColumns(),this,"matrix entry assignment"));
    m(reversed ? k-1-i : i,reversed ? l-1-j : j)=
      force<int_value>(execution_stack.back().get())->int_val();
      // assign |int| from un-popped top
    if (lev==no_value)
      execution_stack.pop_back(); // pop it anyway if result not needed
  }
  break;
    case subscr_base::matrix_column:
  
  { unsigned int j=(index->eval(),get<int_value>()->int_val());
    int_Matrix& m=force<matrix_value>(loc)->val;
  const int_Vector& v=force<vector_value>(execution_stack.back().get())->val;
      // don't pop
    size_t l=m.numColumns();
    if (j>=l)
      throw runtime_error(
        range_mess(j,m.numColumns(),this,"matrix column assignment"));
    if (v.size()!=m.numRows())
      throw runtime_error
        (std::string("Cannot replace column of size ")+str(m.numRows())+
         " by one of size "+str(v.size()));
    m.set_column(reversed ? l-j-1 : j,v);
      // copy value of |int_Vector| into the matrix
    if (lev==no_value)
      execution_stack.pop_back(); // pop the vector if result not needed
  }
  break;
  default: {} // remaining cases are eliminated in type analysis
  }
}

void field_assignment::assign (level lev,shared_value& tupple) const
{ shared_value& field=force<tuple_value>(uniquify(tupple))->val[position];
  rhs->eval();
  push_expanded(lev,field=pop_value());
}

template <bool reversed>
local_component_assignment<reversed>::local_component_assignment
 (id_type arr, expression_ptr&& i,size_t d, size_t o, expression_ptr&& r,
  subscr_base::sub_type k)
: base(arr,std::move(i),std::move(r)), kind(k), depth(d), offset(o) {}

local_field_assignment::local_field_assignment 
  (id_type a, unsigned pos,size_t d, size_t o, expression_ptr&& r)
: field_assignment(a,pos,std::move(r)), depth(d), offset(o) {}

template <bool reversed>
void local_component_assignment<reversed>::evaluate(expression_base::level l)
  const
{ base::assign (l,frame::current->elem(depth,offset),kind); }

void local_field_assignment::evaluate(expression_base::level l) const
{ assign (l,frame::current->elem(depth,offset)); }

} }

